from .base_pipeline import (
    AbstractEnsembleOptimizationPipeline as AbstractEnsembleOptimizationPipeline,
)
from .pipeline_with_openmm import (
    EnsembleOptimizationPipeline as EnsembleOptimizationPipeline,
)



from jaxtyping import Array, Float


def compute_harmonic_steering_force(
    positions: Float[Array, "n_atoms 3"],
    reference_positions: Float[Array, "n_atoms 3"],
    force_constant: Float,
):
    return -force_constant * (positions - reference_positions)



from .common_functions import (
    compute_optimal_scale_and_offset as compute_optimal_scale_and_offset,
)
from .ensemble_likelihood import (
    AbstractImagesToEnsembleLikelihoodFn as AbstractImagesToEnsembleLikelihoodFn,
    ImagesToEnsembleLikelihoodFn as ImagesToEnsembleLikelihoodFn,
)
from .make_model_utils import (
    make_image_model_from_gmm as make_image_model_from_gmm,
)
from .single_likelihood import (
    AbstractImageToWalkerLogLikelihoodFn as AbstractImageToWalkerLogLikelihoodFn,
    GaussianWhiteLogLikelihoodFn as GaussianWhiteLogLikelihoodFn,
    MargGaussianWhiteLogLikelihoodFn as MargGaussianWhiteLogLikelihoodFn,
)



from abc import abstractmethod
from typing import Any

import equinox as eqx
from jaxtyping import Array, Float


class AbstractPriorProjector(eqx.Module, strict=True):
    """
    Abstract class for prior projectors.
    """

    @abstractmethod
    def initialize(
        self,
        init_state: Any | None = None,
    ) -> Any:
        raise NotImplementedError(
            "Abstract method not implemented. "
            "Please implement the initialize method in the subclass."
        )

    @abstractmethod
    def __call__(
        self,
        ref_positions: Float[Array, "n_atoms 3"] | Float[Array, "n_walkers n_atoms 3"],
        state: Any,
        bias_constant: Float,
    ) -> tuple[Float[Array, "n_atoms 3"] | Float[Array, "n_walkers n_atoms 3"], Any]:
        raise NotImplementedError("Abstract method not implemented.")


class AbstractEnsemblePriorProjector(eqx.Module, strict=True):
    """
    Abstract class for ensemble prior projectors.
    """

    projectors: eqx.AbstractVar[list[AbstractPriorProjector]]

    def initialize(
        self,
        init_states: list[Any] | None = None,
    ) -> list[Any]:
        if init_states is None:
            init_states = [None] * len(self.projectors)
        states = []
        for i, projector in enumerate(self.projectors):
            states.append(projector.initialize(init_states[i]))
        return states

    @abstractmethod
    def __call__(
        self,
        ref_positions: Float[Array, "n_walkers n_atoms 3"],
        states: list[Any],
        bias_constant: Float,
    ) -> tuple[Float[Array, "n_walkers n_atoms 3"], list[Any]]:
        raise NotImplementedError("Abstract method not implemented.")



import cryojax.simulator as cxs
import equinox as eqx
from cryospax import RelionParticleDataset, RelionParticleParameterFile

from cryojax_eo.ensemble_optimization import HierarchicalSO3GridSearch


def test_global_SO3_hier_search(
    sample_path_to_pdb1,
    sample_path_to_relion_project,
    sample_path_to_starfile,
):
    gmm_volume = cxs.load_tabulated_volume(
        sample_path_to_pdb1,
        output_type=cxs.GaussianMixtureVolume,
        include_b_factors=True,
        selection_string="not element H",
    )

    relion_dataset = RelionParticleDataset(
        RelionParticleParameterFile(
            sample_path_to_starfile, options=dict(broadcasts_image_config=True)
        ),
        sample_path_to_relion_project,
    )
    pose_search = HierarchicalSO3GridSearch(base_grid_res=1, n_rounds=5, n_candidates=40)
    stack = relion_dataset[0]
    pose = pose_search(
        gmm_volume,
        stack["images"],
        stack["parameters"]["image_config"],
        stack["parameters"]["transfer_theory"],
    )
    eqx.tree_equal(pose, stack["parameters"]["pose"], rtol=1e-2)

    return



import logging
import os
from functools import partial
from typing import Literal

import cryojax.simulator as cxs
import equinox as eqx
import jax
import jax.numpy as jnp
from cryojax.ndimage import CircularCosineMask, FourierGaussian
from cryojax.rotations import SO3
from cryospax import (
    RelionParticleDataset,
    RelionParticleParameterFile,
    simulate_particle_stack,
)
from jaxtyping import Array, Float, PRNGKeyArray

from ..internal._config_validators import DatasetSimulatorConfig
from ..simulator._image_simulation import simulate_image_with_white_gaussian_noise


def make_relion_parameter_file(
    key: PRNGKeyArray, config: DatasetSimulatorConfig
) -> RelionParticleParameterFile:
    """

    This functions writes to disk a starfile containing the
    generated particle parameters. The starfile is saved to the path
    specified in the configuration. The function also returns a
    `cryospax.RelionParticleParameterFile` object that can be used to read the
    starfile and access the particle parameters.

    **Arguments:**
        key: JAX random key for generating random numbers.
        config: Configuration object containing the parameters for generating the dataset.
    **Returns:**
        parameter_file: A RelionParticleParameterFile object containing
        the generated particle parameters.
    """

    config_dict = dict(config.model_dump())
    logging.info("Generating Starfile...")
    key, *subkeys = jax.random.split(key, config_dict["number_of_images"] + 1)
    particle_parameters = _make_particle_parameters(jnp.array(subkeys), config_dict)

    logging.info(
        "Starfile generated. Saved to {}".format(config_dict["path_to_starfile"])
    )

    parameter_file = RelionParticleParameterFile(
        path_to_starfile=config_dict["path_to_starfile"],
        mode="w",
        exist_ok=config_dict["overwrite"],
        options=dict(
            broadcasts_image_config=True
        ),  # Change to False when cryospax is updated
    )
    parameter_file.append(particle_parameters)

    return parameter_file


def simulate_relion_dataset(config: DatasetSimulatorConfig) -> RelionParticleDataset:
    os.makedirs(config.path_to_relion_project, exist_ok=True)

    seed = config.rng_seed
    key = jax.random.key(seed)

    key_param, key_stack = jax.random.split(key)
    parameter_file = make_relion_parameter_file(key_param, config)

    # dumping so serialization happens
    config_dict = dict(config.model_dump())
    volumes = tuple(
        [
            cxs.load_tabulated_volume(
                filename,
                output_type=cxs.GaussianMixtureVolume,
                include_b_factors=config_dict["atomic_models_params"]["loads_b_factors"],
                selection_string=config_dict["atomic_models_params"]["atom_selection"],
            )
            for filename in config_dict["atomic_models_params"]["path_to_atomic_models"]
        ]
    )

    mask = CircularCosineMask(
        coordinate_grid=parameter_file[0]["image_config"].coordinate_grid_in_pixels,
        radius=config_dict["mask_radius"],
        rolloff_width=config_dict["mask_rolloff_width"],
    )

    return _simulate_relion_dataset(
        key=key_stack,
        parameter_file=parameter_file,
        path_to_relion_project=config_dict["path_to_relion_project"],
        images_per_file=config_dict["images_per_file"],
        volumes=volumes,
        ensemble_probabilities=config_dict["atomic_models_params"][
            "atomic_models_probabilities"
        ],
        mask=mask,
        noise_snr_range=config_dict["noise_snr"],
        data_sign=config_dict["data_sign"],
        overwrite=config_dict["overwrite"],
        batch_size=config_dict["batch_size_for_generation"],
    )


def _simulate_relion_dataset(
    key: PRNGKeyArray,
    parameter_file: RelionParticleParameterFile,
    path_to_relion_project: str,
    images_per_file: int,
    volumes: tuple[cxs.GaussianMixtureVolume, ...],
    ensemble_probabilities: Float[Array, " n_volumes"],
    mask: CircularCosineMask,
    noise_snr_range: list[Float],
    data_sign: Literal["dark-on-light", "light-on-dark"],
    *,
    overwrite: bool = False,
    batch_size: int = 1,
) -> RelionParticleDataset:
    assert len(volumes) == len(ensemble_probabilities), (
        "The number of volumes must be equal to the number of ensemble probabilities."
        f" {len(volumes)} != {len(ensemble_probabilities)}"
    )

    # generate random keys
    key_snr, key_noise, key_volumes = jax.random.split(key, 3)

    # Make sure the ensemble probabilities sum to 1
    ensemble_probabilities = jnp.array(ensemble_probabilities)
    ensemble_probabilities /= jnp.sum(ensemble_probabilities)

    # Generate parameters for each image
    snr_per_image = jax.random.uniform(
        key_snr,
        (len(parameter_file),),
        minval=noise_snr_range[0],
        maxval=noise_snr_range[1],
    )

    keys_per_image = jax.random.split(key_noise, len(parameter_file))

    ensemble_indices_per_image = jax.random.choice(
        key_volumes,
        a=len(ensemble_probabilities),
        shape=(len(parameter_file),),
        p=ensemble_probabilities,
        replace=True,
    )

    # Write metadata to disk
    jnp.savez(
        os.path.join(path_to_relion_project, "metadata.npz"),
        snr_per_image=snr_per_image,
        ensemble_indices_per_image=ensemble_indices_per_image,
    )
    logging.info("Metadata: noise variance and ensemble indices saved to:")
    logging.info(f"  {os.path.join(path_to_relion_project, 'metadata.npz')}")

    # Bundle arguments and write images
    data_sign_factor = -1.0 if data_sign == "dark-on-light" else 1.0
    constant_args = (volumes, mask, data_sign_factor)
    per_particle_args = (
        keys_per_image,
        ensemble_indices_per_image,
        snr_per_image,
    )

    relion_dataset = RelionParticleDataset(
        parameter_file=parameter_file,
        path_to_relion_project=path_to_relion_project,
        mode="w",
        mrcfile_settings={"overwrite": overwrite},
    )

    simulate_particle_stack(
        dataset=relion_dataset,
        simulate_fn=simulate_image_with_white_gaussian_noise,
        constant_args=constant_args,
        per_particle_args=per_particle_args,
        images_per_file=images_per_file,
        batch_size=batch_size,
        overwrite=overwrite,
    )
    logging.info(f"Images generated. Saved to {path_to_relion_project}")
    logging.info("Simulated dataset generation complete.")
    return relion_dataset


@partial(eqx.filter_vmap, in_axes=(0, None))
def _make_particle_parameters(key: PRNGKeyArray, config: dict) -> dict:
    """
    WARNING: this function assumes the `config` has been validated
    by `cryojax_ensemble_refinement.internal.GeneratorConfig`.
    Skipping this step could lead to unexpected behavior.
    """
    image_config = cxs.BasicImageConfig(
        shape=(config["box_size"], config["box_size"]),
        pixel_size=config["pixel_size"],
        voltage_in_kilovolts=config["voltage_in_kilovolts"],
    )
    # Generate random parameters

    # Pose
    # ... instantiate rotations
    key, subkey = jax.random.split(key)  # split the key to use for the next random number

    rotation = SO3.sample_uniform(subkey)
    key, subkey = jax.random.split(key)  # do this everytime you use a key!!

    # ... now in-plane translation
    offset_x_in_angstroms = jax.random.uniform(
        subkey,
        (1,),
        minval=config["offset_x_in_angstroms"][0],
        maxval=config["offset_x_in_angstroms"][1],
    )
    key, subkey = jax.random.split(key)

    offset_y_in_angstroms = jax.random.uniform(
        subkey,
        (1,),
        minval=config["offset_y_in_angstroms"][0],
        maxval=config["offset_y_in_angstroms"][1],
    )
    key, subkey = jax.random.split(key)

    in_plane_offset_in_angstroms = jnp.concatenate(
        [offset_x_in_angstroms, offset_y_in_angstroms]
    )
    # ... convert 2D in-plane translation to 3D, setting the out-of-plane translation to
    # zero
    offset_in_angstroms = jnp.pad(in_plane_offset_in_angstroms, ((0, 1),))
    # ... build the pose
    pose = cxs.EulerAnglePose.from_rotation_and_translation(rotation, offset_in_angstroms)

    # CTF Parameters
    # ... defocus
    defocus_in_angstroms = jax.random.uniform(
        subkey,
        (),
        minval=config["defocus_in_angstroms"][0],
        maxval=config["defocus_in_angstroms"][1],
    )
    key, subkey = jax.random.split(key)

    astigmatism_in_angstroms = jax.random.uniform(
        subkey,
        (),
        minval=config["astigmatism_in_angstroms"][0],
        maxval=config["astigmatism_in_angstroms"][1],
    )
    key, subkey = jax.random.split(key)

    astigmatism_angle = jax.random.uniform(
        subkey,
        (),
        minval=config["astigmatism_angle_in_degrees"][0],
        maxval=config["astigmatism_angle_in_degrees"][1],
    )
    key, subkey = jax.random.split(key)

    phase_shift = jax.random.uniform(
        subkey, (), minval=config["phase_shift"][0], maxval=config["phase_shift"][1]
    )
    key, subkey = jax.random.split(key)

    b_factor = jax.random.uniform(
        subkey,
        (),
        minval=config["envelope_b_factor"][0],
        maxval=config["envelope_b_factor"][1],
    )
    # no more random numbers needed

    # now generate your non-random values
    spherical_aberration_in_mm = config["spherical_aberration_in_mm"]
    amplitude_contrast_ratio = config["amplitude_contrast_ratio"]
    ctf_scale_factor = config["ctf_scale_factor"]

    b_factor = jnp.clip(b_factor, 1e-16, None)
    envelope = FourierGaussian(b_factor=b_factor, amplitude=ctf_scale_factor)

    # ... build the CTF
    transfer_theory = cxs.ContrastTransferTheory(
        ctf=cxs.AstigmaticCTF(
            defocus_in_angstroms=defocus_in_angstroms,
            astigmatism_in_angstroms=astigmatism_in_angstroms,
            astigmatism_angle=astigmatism_angle,
            spherical_aberration_in_mm=spherical_aberration_in_mm,
        ),
        amplitude_contrast_ratio=amplitude_contrast_ratio,
        phase_shift=phase_shift,
        envelope=envelope,
    )

    relion_particle_parameters = {
        "image_config": image_config,
        "pose": pose,
        "transfer_theory": transfer_theory,
    }
    return relion_particle_parameters



import jax.numpy as jnp
from jaxtyping import Array, Float


def rigid_align_positions(
    target_pos: Float[Array, "n_atoms 3"], ref_pos: Float[Array, "n_atoms 3"]
) -> tuple[Float[Array, "n_atoms 3"], Float[Array, "3 3"], Float[Array, "1 3"]]:
    com_ref = jnp.mean(ref_pos, axis=0, keepdims=True)
    com_target = jnp.mean(target_pos, axis=0, keepdims=True)

    cross_cov_matrix = jnp.dot((ref_pos - com_ref).T, target_pos - com_target)

    U, _, Vh = jnp.linalg.svd(cross_cov_matrix)
    det = jnp.linalg.det(U) * jnp.linalg.det(Vh)
    rot_matrix = U @ jnp.diag(jnp.array([1.0, 1.0, det])) @ Vh

    displacement = com_ref - com_target @ rot_matrix.T

    aligned_pos = target_pos @ rot_matrix.T + displacement
    return aligned_pos, rot_matrix, displacement



"""
Weight and position optimizers for ensemble refinement.
"""

import equinox as eqx
import jax
import jax.numpy as jnp
from jaxtyping import Array, Float, Int

from .model_to_volume_loss import ModelToVolumeLikelihoodFn


class SteepestDescWalkerFlexibleFitting(eqx.Module):
    step_size: Float
    n_steps: Int
    model_to_vol_likelihood_fn: ModelToVolumeLikelihoodFn

    def __init__(
        self,
        n_steps: Int,
        step_size: Float,
        model_to_vol_likelihood_fn: ModelToVolumeLikelihoodFn,
    ):
        assert n_steps > 0, "n_steps must be positive"
        assert step_size >= 0, "step_size must be non-negative."

        self.n_steps = n_steps
        self.model_to_vol_likelihood_fn = model_to_vol_likelihood_fn
        self.step_size = step_size

    def __call__(
        self,
        walkers,
        reference_volume,
    ) -> tuple[Float, Float[Array, "n_atoms 3"]]:
        loss = jnp.inf
        for _ in range(self.n_steps):
            loss, walkers = _optimize_walkers_positions(
                walkers,
                reference_volume,
                self.step_size,
                self.model_to_vol_likelihood_fn,
            )

        return loss, walkers


@eqx.filter_jit
def _optimize_walkers_positions(
    walkers: Float[Array, "n_atoms 3"],
    reference_volume: Float[Array, "n_pixels n_pixels n_pixels"],
    step_size: Float,
    likelihood_fn: ModelToVolumeLikelihoodFn,
) -> tuple[Float, Float[Array, "n_atoms 3"]]:
    def _lklhood_fn(walker, ref_volume):
        return likelihood_fn(walker, ref_volume)

    loss, gradients = jax.value_and_grad(
        _lklhood_fn,
        argnums=0,
    )(
        walkers,
        reference_volume,
    )

    norms = jnp.linalg.norm(gradients, axis=(1), keepdims=True)
    # set small norms to 1 (avoid making small gradients large!)
    norms = jnp.where(norms < 1e-12, 1.0, norms)
    gradients /= norms

    return loss, walkers - step_size * gradients



import os

import jax
import pytest


jax.config.update("jax_enable_x64", True)


@pytest.fixture
def sample_path_to_pdb1():
    return os.path.join(os.path.dirname(__file__), "data", "ala_model_0.pdb")


@pytest.fixture
def sample_path_to_pdb2():
    return os.path.join(os.path.dirname(__file__), "data", "ala_model_1.pdb")


@pytest.fixture
def sample_path_gmm_model():
    return os.path.join(os.path.dirname(__file__), "data", "test_groel_gmm_model.npz")


@pytest.fixture
def sample_path_mrc_file():
    return os.path.join(os.path.dirname(__file__), "data", "test_groel_volume.mrc")


@pytest.fixture
def sample_path_groel_pdb():
    return os.path.join(os.path.dirname(__file__), "data", "apo_groel.pdb")


@pytest.fixture
def sample_path_to_starfile():
    return os.path.join(
        os.path.dirname(__file__), "data", "particle_stack/test_starfile.star"
    )


@pytest.fixture
def sample_path_to_relion_project():
    return os.path.join(os.path.dirname(__file__), "data", "particle_stack/")



"""
Base optimizer object for ensemble refinement.
"""

from abc import abstractmethod

from equinox import AbstractVar, Module
from jax_dataloader import DataLoader
from jaxtyping import Array, Float

from .._pose_search import HierarchicalSO3GridSearch
from ._loss_functions import AbstractImagesToEnsembleLikelihoodFn


class AbstractEnsembleParameterOptimizer(Module):
    """Abstract interface for objects that optimize parameters
    of an ensemble of structures.
    """

    n_steps: AbstractVar[int]
    ensemble_likelihood_fn: AbstractVar[AbstractImagesToEnsembleLikelihoodFn]
    pose_search: AbstractVar[HierarchicalSO3GridSearch | None]

    @abstractmethod
    def __call__(
        self,
        walkers: Float[Array, "n_walkers n_atoms 3"],
        weights: Float[Array, " n_walkers"],
        dataloader: DataLoader,
    ) -> Float:
        raise NotImplementedError



import jax.numpy as jnp
from jaxtyping import Array, Float


def compute_optimal_scale_and_offset(
    target_image: Float[Array, "n_pixels n_pixels"],
    ref_image: Float[Array, "n_pixels n_pixels"],
    signal_region: Float[Array, "n_pixels n_pixels"] | None = None,
) -> tuple[Float, Float]:
    if signal_region is None:
        signal_region = jnp.ones_like(target_image, dtype=bool)

    cc = jnp.mean(target_image**2, where=signal_region)
    co = jnp.mean(ref_image * target_image, where=signal_region)
    c = jnp.mean(target_image, where=signal_region)
    o = jnp.mean(ref_image, where=signal_region)

    scale = (co - c * o) / (cc - c**2)
    offset = o - scale * c

    return scale, offset



import jax.numpy as jnp
import mdtraj
import numpy as np
from cryojax.constants import (
    PengScatteringFactorParameters,
    b_factor_to_variance,
)
from cryojax.io import read_atoms_from_pdb
from jaxtyping import Array, Float, Int

from cryojax_eo.utils import rigid_align_positions


def read_walkers_from_pdbs(
    filenames: list[str],
    *,
    loads_b_factors: bool = False,
) -> tuple[
    Float[Array, "n_walkers n_atoms 3"],
    Float[Array, "n_atoms n_gaussians"],
    Float[Array, "n_atoms n_gaussians"],
]:
    """
    Reads atomic models from PDB files and extracts positions, scattering amplitudes, and
    variances for each model. The function can also load Debye-Waller b-factors
    if specified. Due to the constraints of this library, the variances and
    amplitudes must be consistent across all models. This means that the atomic models
    must have the same atomic composition, but the positions can differ.

    **Arguments:**
    - `filenames`:
        List of paths to PDB files containing atomic models.
    - `loads_b_factors`:
        If `True`, the function will read the B-factors from the PDB files
        and use them to compute the variances. If `False`, default variances
        based on atomic types will be used.

    **Returns:**
    A tuple containing:
    - `walkers`:
        A JAX array of shape `(n_walkers, n_atoms, 3)`
        with the atomic positions for each model.
    - `amplitudes`:
        A JAX array of shape `(n_atoms, n_gaussians)` with the scattering amplitudes.
    - `variances`:
        A JAX array of shape `(n_atoms, n_gaussians)` with the scattering variances.
    """

    walkers_and_params = {}
    for i in range(len(filenames)):
        atom_positions, atomic_numbers, atom_properties = read_atoms_from_pdb(
            filenames[i], center=True, loads_properties=True
        )
        variances, amplitudes = _compute_scattering_params(
            atomic_numbers,
            atom_properties["b_factors"] if loads_b_factors else None,
        )
        walkers_and_params[i] = {
            "positions": atom_positions,
            "amplitudes": amplitudes,
            "variances": variances,
        }

    _assert_consistent_scattering_params(walkers_and_params)

    variances = walkers_and_params[0]["variances"]
    amplitudes = walkers_and_params[0]["amplitudes"]
    walkers = np.array(
        [walkers_and_params[i]["positions"] for i in range(len(filenames))]
    )
    topology = mdtraj.load(filenames[0]).topology
    walkers = _align_walkers_to_reference(walkers, topology)

    return walkers, amplitudes, variances


def _align_walkers_to_reference(
    walkers: Float[Array, "n_walkers n_atoms 3"],
    topology: mdtraj.Topology,
):
    aligned_walkers = np.zeros_like(walkers)
    atom_indices = topology.select("name CA")
    for i in range(walkers.shape[0]):
        _, rot_matrix, displacement = rigid_align_positions(
            walkers[i, atom_indices], walkers[0, atom_indices]
        )
        aligned_walkers[i] = walkers[i] @ rot_matrix.T + displacement

    return aligned_walkers


def _assert_consistent_scattering_params(walkers_and_params: dict):
    """
    Asserts that the scattering amplitudes and variances are consistent across
    all models. This is necessary because the current implementation of the optimization
    process assumes that all models have the same atomic composition,
    and thus the same scattering parameters.
    """
    reference_amplitudes = walkers_and_params[0]["amplitudes"]
    reference_variances = walkers_and_params[0]["variances"]

    for i, (model_id, params) in enumerate(walkers_and_params.items()):
        if i == 0:
            continue  # Skip the reference model

        _var_shapes_match = params["amplitudes"].shape == reference_amplitudes.shape
        _amp_shapes_match = params["variances"].shape == reference_variances.shape
        if not _var_shapes_match or not _amp_shapes_match:
            raise ValueError(
                f"Scattering parameters for model {model_id} "
                "have inconsistent shapes with the reference model."
            )
        else:
            if not jnp.allclose(params["amplitudes"], reference_amplitudes):
                raise ValueError(
                    f"Scattering amplitudes for model {model_id} "
                    "are inconsistent with the reference model."
                )
            if not jnp.allclose(params["variances"], reference_variances):
                raise ValueError(
                    f"Scattering variances for model {model_id} "
                    "are inconsistent with the reference model."
                )


def _compute_scattering_params(
    atomic_numbers: Int[Array, " n_atoms"],
    b_factors: Float[Array, " n_atoms"] | None,
):
    scattering_factors = PengScatteringFactorParameters(atomic_numbers)
    amplitudes = scattering_factors.a

    if b_factors is None:
        variances = scattering_factors.b
    else:
        variances = b_factor_to_variance(scattering_factors.b + b_factors[:, None])

    return amplitudes, variances



import glob
import os
from pathlib import Path

from natsort import natsorted
from pydantic import (
    DirectoryPath,
    FilePath,
)


def _validate_file_with_type(filename: str, file_type: str) -> str:
    """
    Check if the file is a PDB file.
    """
    assert Path(filename).exists(), f"File {filename} does not exist."
    assert Path(filename).is_file(), f"Path {filename} is not a file."

    if not Path(filename).suffix == f"{file_type}":
        raise ValueError(f"File {filename} is not a {file_type} file.")
    return filename


# Might be useful, and I don't want to figure it out again
# TODO: remove if not needed
def _contains_file_type(
    directory_path: DirectoryPath, file_type: str | list[str]
) -> DirectoryPath:
    if isinstance(file_type, str):
        file_type = [file_type]

    failing_types = []
    for ftype in file_type:
        files_in_directory = glob.glob(os.path.join(directory_path, f"*.{ftype}"))
        if len(files_in_directory) == 0:
            failing_types.append(ftype)

    if len(failing_types) > 0:
        raise ValueError(
            f"Directory {directory_path} does not contain any files "
            + f"of type(s): {', '.join(failing_types)}"
        )

    return directory_path


def _validate_files_with_type(
    path_to_files: str | list[FilePath], file_types: list[str]
) -> list[str]:
    if isinstance(path_to_files, str):
        if "*" in path_to_files:
            output = [Path(f) for f in natsorted(glob.glob(path_to_files))]
        elif Path(path_to_files).is_file():
            output = [Path(path_to_files)]
        else:
            raise ValueError(
                f"Path {path_to_files} is not a file or does not use * wild card."
            )
    elif isinstance(path_to_files, list):
        output = [Path(f) for f in path_to_files]

    for f in output:
        assert f.exists(), f"{f} does not exist."
        assert f.is_file(), f"{f} is not a file."
        assert f.suffix in file_types, (
            f"{f} is not a valid file type. "
            + f"Valid file types are: {', '.join([f'.{file_types}'])}"
        )
    return [str(f) for f in output]



import pathlib
from abc import abstractmethod
from typing import Any

import optax
from equinox import AbstractVar, Module
from jax_dataloader import DataLoader
from jaxtyping import Array, Float, Int, PRNGKeyArray

from .._likelihood_optimization.base_optimizer import AbstractEnsembleParameterOptimizer
from .._prior_projection.base_prior_projector import AbstractEnsemblePriorProjector


class AbstractEnsembleOptimizationPipeline(Module, strict=True):
    """
    Abstract class for ensemble refinement pipelines.
    """

    prior_projector: AbstractVar[AbstractEnsemblePriorProjector]
    likelihood_optimizer: AbstractVar[AbstractEnsembleParameterOptimizer]
    n_steps: AbstractVar[Int]

    @abstractmethod
    def run(
        self,
        key: PRNGKeyArray,
        initial_walkers: Float[Array, "n_walkers n_atoms 3"],
        initial_weights: Float[Array, " n_walkers"],
        dataloader: DataLoader,
        bias_constant_scheduler: optax.Schedule,
        *,
        output_directory: str | pathlib.Path,
        initial_state_for_projector: Any = None,
    ) -> tuple[
        Float[Array, "n_steps n_walkers n_atoms 3"],
        Float[Array, "n_steps n_walkers"],
    ]:
        raise NotImplementedError



from pathlib import Path
from typing import Literal, overload

import yaml

from ._config_validators import DatasetSimulatorConfig, EnsOptMDConfig, GMMFitConfig


@overload
def load_config(
    path_to_config: str | Path,
    config_mode: Literal["data simulation"],
) -> DatasetSimulatorConfig: ...


@overload
def load_config(
    path_to_config: str | Path,
    config_mode: Literal["ensemble optimization"],
) -> EnsOptMDConfig: ...


@overload
def load_config(
    path_to_config: str | Path,
    config_mode: Literal["gmm fitting"],
) -> GMMFitConfig: ...


def load_config(
    path_to_config: str | Path,
    config_mode: Literal["data simulation", "ensemble optimization", "gmm fitting"],
) -> DatasetSimulatorConfig | EnsOptMDConfig | GMMFitConfig:
    """
    Load a configuration file and parse it into the appropriate configuration object.

    Parameters
    ----------
    path_to_config : str | Path
        Path to the configuration file (YAML format).
    config_mode : Literal["data simulation", "ensemble optimization", "gmm fitting"]
        Type of configuration to load. Must be one of "data simulation",
        "ensemble optimization", or "gmm fitting".

    Returns
    -------
    DatasetSimulatorConfig | EnsOptMDConfig | GMMFitConfig
        Parsed configuration object.

    Raises
    ------
    ValueError
        If the provided config_mode is not recognized.
    """
    with open(path_to_config) as f:
        config_dict = yaml.safe_load(f)

    if config_mode == "data simulation":
        return DatasetSimulatorConfig(**config_dict)
    elif config_mode == "ensemble optimization":
        return EnsOptMDConfig(**config_dict)
    elif config_mode == "gmm fitting":
        return GMMFitConfig(**config_dict)
    else:
        raise ValueError(
            f"Unknown config_mode: {config_mode}. Must be one of"
            + " 'data simulation', 'ensemble optimization', or 'gmm fitting'."
        )



"""
Base class for force fields used in custom prior projection methods.
"""

from abc import abstractmethod
from collections.abc import Callable
from typing import TypeVar

import equinox as eqx
from jaxtyping import Array, Float


EnergyFuncArgs = TypeVar("EnergyFuncArgs")


class AbstractForceField(eqx.Module, strict=True):
    """
    Abstract base class for a force field.
    """

    energy_fn: Callable[[Float[Array, "n_atoms 3"], EnergyFuncArgs], Float]
    energy_fn_args: EnergyFuncArgs

    @abstractmethod
    def __call__(
        self,
        positions: Float[Array, "n_atoms 3"],
    ):
        """
        Compute the energy of the system.
        """
        raise NotImplementedError



import cryojax.simulator as cxs
import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array, Float, Int


class ModelToVolumeLikelihoodFn(eqx.Module):
    variances: Float[Array, "n_atoms n_gaussians_per_atom"]
    amplitudes: Float[Array, "n_atoms n_gaussians_per_atom"]
    voxel_size: Float
    batch_size_for_z_planes: Int
    n_batches_of_atoms: Int

    def __init__(
        self,
        amplitudes: Float[Array, "n_atoms n_gaussians_per_atom"],
        variances: Float[Array, "n_atoms n_gaussians_per_atom"],
        voxel_size: Float,
        *,
        batch_size_for_z_planes: Int = 1,
        n_batches_of_atoms: Int = 1,
    ):
        assert (amplitudes > 0).all(), "Amplitudes must be positive."
        assert (variances > 0).all(), "Variances must be positive."
        assert voxel_size > 0, "Voxel size must be positive."
        assert n_batches_of_atoms > 0, "n_batches_of_atoms must be positive."
        assert batch_size_for_z_planes > 0, "batch_size_for_z_planes must be positive."

        self.variances = variances
        self.amplitudes = amplitudes

        self.voxel_size = voxel_size
        self.batch_size_for_z_planes = int(batch_size_for_z_planes)
        self.n_batches_of_atoms = int(n_batches_of_atoms)

    def __call__(
        self,
        walker: Float[Array, "n_atoms 3"],
        reference_volume: Float[Array, "dim dim dim"],
    ) -> Float:
        # Compute the model-to-volume loss
        return 1 - _model_to_volume_crosscorrelation(
            walker,
            self.amplitudes,
            self.variances,
            reference_volume,
            self.voxel_size,
            batch_size_for_z_planes=self.batch_size_for_z_planes,
            n_batches_of_atoms=self.n_batches_of_atoms,
        )


def _model_to_volume_crosscorrelation(
    walker: Float[Array, "n_atoms 3"],
    amplitudes: Float[Array, "n_atoms n_gaussians_per_atom"],
    variances: Float[Array, "n_atoms n_gaussians_per_atom"],
    reference_volume: Float[Array, "dim dim dim"],
    voxel_size: Float,
    *,
    batch_size_for_z_planes: int = 1,
    n_batches_of_atoms: int = 1,
) -> Float:
    volume_shape = reference_volume.shape

    comp_volume = cxs.GaussianMixtureVolume(
        walker,
        amplitudes,
        variances,
    ).to_real_voxel_grid(
        volume_shape,  # type: ignore
        voxel_size,
        batch_options={
            "batch_size": batch_size_for_z_planes,
            "n_batches": n_batches_of_atoms,
        },
    )

    cross_correlation = (
        jnp.sum(comp_volume * reference_volume)
        / jnp.linalg.norm(comp_volume)
        / jnp.linalg.norm(reference_volume)
    )

    return cross_correlation



from typing import Any

import cryojax.simulator as cxs
import equinox as eqx
import jax
import jax.numpy as jnp
from jaxtyping import Array, Float

from .geometry import (
    getbestneighbors_base_SO3,
    getbestneighbors_next_SO3,
    grid_SO3,
)
from .pose_offset import compute_correlation_at_optimal_offset


@eqx.filter_vmap(in_axes=(0, None, None, None, None))
def _loss_for_grid_search(
    quat: Float[Array, "4"],
    volume: cxs.AbstractVolumeRepresentation,
    target_image: Float[Array, "H W"],
    image_config: cxs.BasicImageConfig,
    transfer_theory: cxs.ContrastTransferTheory,
) -> tuple[Float[Array, ""], Float[Array, " 2"]]:
    """
    Computes the correlation at the optimal shift using Plancherel's theorem and
    Fourier's convolution theorem for a given quaternion.

    The function outputs the negative correlation as this is a loss function.
    """
    pose = cxs.QuaternionPose(
        offset_x_in_angstroms=0.0, offset_y_in_angstroms=0.0, wxyz=quat
    )

    computed_image_no_shift = cxs.make_image_model(
        volume,
        image_config,
        pose,
        transfer_theory,
        normalizes_signal=True,
    ).simulate()

    # get the optimal shift
    correlation, optimal_offset = compute_correlation_at_optimal_offset(
        target_image,
        computed_image_no_shift,
        image_config.coordinate_grid_in_angstroms,
    )

    return -correlation, optimal_offset


class _HierSO3GriSearchCarry(eqx.Module):
    losses: Float[Array, " N"]
    quats: Float[Array, "N 4"]
    offsets: Float[Array, "N 2"]
    grid_indices: Any
    curr_resolution: int


class HierarchicalSO3GridSearch(eqx.Module):
    """
    Perform a hierarchical grid search over SO3 to find the optimal pose
    using a shift-invariant metric based on correlation. Also provides
    the optimal shift for the pose found.

    **Attributes:**
    - `base_quats`:
        Base quaternions for the SO3 grid at the coarsest resolution.
    - `n_rounds`:
        Number of rounds to perform the hierarchical search.
    - `n_candidates`:
        Number of candidate quaternions to consider in each round.
    - `base_grid_res`:
        Base grid resolution for the SO3 grid.
    """

    base_quats: Float[Array, "N 4"]
    n_rounds: int
    n_candidates: int
    base_grid_res: int

    def __init__(self, base_grid_res, n_rounds, n_candidates):
        """
        Initialize the HierarchicalSO3GridSearch.

        **Arguments:**
        - `base_grid_res`:
            Base grid resolution for the SO3 grid.
        - `n_rounds`:
            Number of rounds to perform the hierarchical search.
        - `n_candidates`:
            Number of candidate quaternions to consider in each round.
        """
        assert base_grid_res >= 1, "Base grid must be at least 1."
        assert n_rounds >= 0, "Number of rounds must be non-negative."
        assert n_candidates >= 1, "Number of candidates must be at least 1."

        self.base_grid_res = base_grid_res
        self.base_quats = grid_SO3(base_grid_res)
        self.n_rounds = n_rounds
        self.n_candidates = n_candidates

    def __call__(
        self,
        volume: cxs.AbstractVolumeRepresentation,
        image: Float[Array, "H W"],
        image_config: cxs.BasicImageConfig,
        transfer_theory: cxs.ContrastTransferTheory,
    ) -> cxs.QuaternionPose:
        """Perform the hierarchical SO3 grid search to find the optimal pose.

        **Arguments:**
        - `volume`:
            `cryojax` `AbstractVolumeRepresentation` representing a volumetric density.
        - `image`:
            Target image to compare against through the search.
        - `image_config`:
            Configuration for image simulation.
        - `transfer_theory`:
            Contrast transfer theory to use for image simulation.

        **Returns:**
        - `optimal_pose`:
            The optimal pose found through the hierarchical SO3 grid search.
        """

        image = jnp.asarray(image)
        losses, offsets = _loss_for_grid_search(
            self.base_quats,
            volume,
            image,
            image_config,
            transfer_theory,
        )

        if self.n_rounds == 0:
            best_index = jnp.argmin(losses)
            optimal_pose = cxs.QuaternionPose(
                wxyz=self.base_quats[best_index],
                offset_x_in_angstroms=offsets[best_index, 0],
                offset_y_in_angstroms=offsets[best_index, 1],
            )

        else:
            # Do the first iteration outside the loop
            quats, grid_indices = getbestneighbors_base_SO3(
                losses,
                self.base_quats,
                N=self.n_candidates,
                base_resol=self.base_grid_res,
            )
            losses, offsets = _loss_for_grid_search(
                quats,
                volume,
                image,
                image_config,
                transfer_theory,
            )
            carry = _HierSO3GriSearchCarry(
                losses=losses,
                quats=quats,
                grid_indices=grid_indices,
                offsets=offsets,
                curr_resolution=self.base_grid_res + 1,
            )
            # Then do the rest of the rounds in a loop
            carry = jax.lax.fori_loop(
                lower=2,
                upper=self.n_rounds,
                body_fun=lambda _, x: _run_global_SO3_step(
                    x, self.n_candidates, volume, image, image_config, transfer_theory
                ),
                init_val=carry,
            )
            best_index = jnp.argmin(carry.losses)
            optimal_pose = cxs.QuaternionPose(
                wxyz=carry.quats[best_index],
                offset_x_in_angstroms=carry.offsets[best_index, 0],
                offset_y_in_angstroms=carry.offsets[best_index, 1],
            )

        return optimal_pose


def _run_global_SO3_step(
    hier_grid_search_carry: _HierSO3GriSearchCarry,
    n_candidates: int,
    volume: cxs.AbstractVolumeRepresentation,
    image: Float[Array, "H W"],
    image_config: cxs.BasicImageConfig,
    transfer_theory: cxs.ContrastTransferTheory,
):
    quats, grid_indices = getbestneighbors_next_SO3(
        hier_grid_search_carry.losses,
        hier_grid_search_carry.quats,
        hier_grid_search_carry.grid_indices,
        N=n_candidates,
    )
    losses, offsets = _loss_for_grid_search(
        quats,
        volume,
        image,
        image_config,
        transfer_theory,
    )
    return _HierSO3GriSearchCarry(
        losses=losses,
        quats=quats,
        offsets=offsets,
        grid_indices=grid_indices,
        curr_resolution=hier_grid_search_carry.curr_resolution + 1,
    )


def local_SO3_hier_search(lossfn, base_grid_res=1, n_rounds=5, n_candidates=40):
    raise NotImplementedError(
        "Local SO3 hierarchical search is not implemented yet. "
        "Please use global SO3 hierarchical search instead."
    )



"""
Functions for running MD simulations using OpenMM.

Functions
---------
run_md_openmm
    Run MD simulations using OpenMM
"""

import os
import pathlib
import shutil
import warnings
from collections.abc import Callable
from functools import partial
from pathlib import Path
from typing_extensions import override

import jax
import jax.numpy as jnp
import mdtraj
import numpy as np
from jaxtyping import Array, Float, Int


try:
    import openmm
    import openmm.app as openmm_app
    import openmm.unit as openmm_unit

    _HAS_OPENMM = True

except ImportError:
    _HAS_OPENMM = False
    warnings.warn(
        "OpenMM is not installed. Please install OpenMM if using any features "
        + "that use molecular dynamics, e.g., ensemble optimization "
        + "or flexible fitting."
    )


from ..base_prior_projector import AbstractEnsemblePriorProjector, AbstractPriorProjector


def _get_default_md_params() -> dict:
    return {
        "forcefield": "amber14-all.xml",
        "water_model": "amber14/tip3p.xml",
        "nonbondedMethod": openmm_app.PME,
        "nonbondedCutoff": 1.0 * openmm_unit.nanometer,
        "constraints": openmm_app.HBonds,
        "temperature": 300.0 * openmm_unit.kelvin,
        "friction": 1.0 / openmm_unit.picosecond,
        "timestep": 0.002 * openmm_unit.picoseconds,
        "platform": "CPU",
        "properties": {"Threads": "1"},
    }


class SteeredMDSimulator(AbstractPriorProjector, strict=True):
    n_steps: Int
    simulation: openmm_app.Simulation
    restrain_atom_list: list[Int]
    base_state_file_path: str

    def __init__(
        self,
        path_to_initial_pdb: str | pathlib.Path,
        n_steps: Int,
        restrain_atom_list: list[Int],
        parameters_for_md: dict,
        base_state_file_path: str,
        *,
        make_simulation_fn: Callable | None = None,
    ):
        if not _HAS_OPENMM:
            raise ImportError(
                "OpenMM is not installed. Please install OpenMM if using any features "
                + "that use molecular dynamics, e.g., the ensemble optimization pipeline "
                + "or flexible fitting."
            )
        pdb = openmm_app.PDBFile(str(path_to_initial_pdb))
        self.restrain_atom_list = restrain_atom_list

        self.base_state_file_path = _validate_base_state_file_path(base_state_file_path)

        self.n_steps = n_steps
        if make_simulation_fn is None:
            parameters_for_md = _validate_and_set_params_for_md(parameters_for_md)
            self.simulation = _default_make_sim_fn(parameters_for_md, pdb.topology)

        else:
            self.simulation = make_simulation_fn(parameters_for_md, pdb.topology)

        self.simulation.context.setPositions(pdb.positions)

    @override
    def initialize(self, init_state: str | None = None) -> str:
        if init_state is not None:
            assert pathlib.Path(init_state).exists(), (
                "init_state does not exist. "
                "Please set to None or provide valid state file."
            )
            self.simulation.loadState(str(init_state))
            path_to_state_file = f"{self.base_state_file_path}0.xml"
            if os.path.exists(path_to_state_file):
                if Path(init_state).samefile(path_to_state_file):
                    Warning(
                        "The provided init_state has the same base name as the "
                        + "base_state_file_path. "
                        + "This may cause overwriting of the state file."
                    )
                    path_to_state_file = f"{self.base_state_file_path}1.xml"

        else:
            path_to_state_file = f"{self.base_state_file_path}0.xml"
            self.simulation.minimizeEnergy()
            self.simulation.step(100)

        self.simulation.saveState(path_to_state_file)

        return path_to_state_file

    @override
    def __call__(
        self, ref_walkers: Float[Array, "n_atoms 3"], state: str, bias_constant: float
    ) -> tuple[Float[Array, "n_atoms 3"], str]:
        _assert_is_valid_state_file(state, self.base_state_file_path)

        simulation = _add_restraint_force_to_simulation(
            self.simulation,
            mdtraj.Trajectory(
                ref_walkers / 10.0, self.simulation.topology
            ).openmm_positions(0),
            self.restrain_atom_list,
            bias_constant,
        )

        # print("Reinitialize")
        simulation.context.reinitialize()

        # print("Loading state")
        simulation.loadState(state)

        # print("Running Simulation")
        simulation.step(self.n_steps)
        positions = simulation.context.getState(getPositions=True).getPositions()
        velocities = simulation.context.getState(getVelocities=True).getVelocities()
        # print("Cleaning up")
        simulation = _remove_last_force_from_simulation(simulation)
        simulation.context.reinitialize()  # preserveState=True)

        simulation.context.setPositions(positions)
        simulation.context.setVelocities(velocities)

        state = _get_next_state_file_path(self.base_state_file_path, state)
        simulation.saveState(state)

        # print("Saved states... Finishing.")

        positions = (
            simulation.context.getState(getPositions=True)
            .getPositions(asNumpy=True)
            .value_in_unit(openmm_unit.angstrom)
        )
        return jnp.array(positions), state


class EnsembleSteeredMDSimulator(AbstractEnsemblePriorProjector, strict=True):
    projectors: list[SteeredMDSimulator]

    def __init__(self, md_simulators: list[SteeredMDSimulator]):
        if not _HAS_OPENMM:
            raise ImportError(
                "OpenMM is not installed. Please install OpenMM if using any features "
                + "that use molecular dynamics, e.g., the ensemble optimization pipeline "
                + "or flexible fitting."
            )
        self.projectors = md_simulators

    @override
    def __call__(
        self,
        ref_positions: Float[Array, "n_walkers n_atoms 3"],
        states: list[str],
        bias_constant: float,
    ) -> tuple[Float[Array, "n_walkers n_atoms 3"], list[str]]:
        projected_walkers = np.zeros_like(ref_positions)
        for i, projector in enumerate(self.projectors):
            projected_walkers[i], states[i] = projector(
                ref_positions[i], states[i], bias_constant
            )
        return jnp.array(projected_walkers), states


def compute_biasing_constant(
    target_percentage: Float,
    path_to_initial_pdb: str,
    positions_for_bias,
    n_steps: Int,
    stride: Int = 1,
    atom_selection: str = "not element H",
    *,
    make_simulation_fn: Callable[[dict, openmm_app.Topology], openmm_app.Simulation]
    | None = None,
    parameters_for_md: dict = {},
):
    if not _HAS_OPENMM:
        raise ImportError(
            "OpenMM is not installed. Please install OpenMM if using any features "
            + "that use molecular dynamics, e.g., the ensemble optimization pipeline "
            + "or flexible fitting."
        )
    if make_simulation_fn is None:
        parameters_for_md = _validate_and_set_params_for_md(parameters_for_md)
        make_simulation_fn = _default_make_sim_fn

    restrain_atom_list = mdtraj.load(str(path_to_initial_pdb)).topology.select(
        atom_selection
    )

    pdb = openmm_app.PDBFile(str(path_to_initial_pdb))
    simulation = make_simulation_fn(parameters_for_md, pdb.topology)
    simulation.context.setPositions(pdb.positions)
    simulation.minimizeEnergy()

    # initial_positions = simulation.context.getState(
    #     getPositions=True
    # ).getPositions()

    # initial_positions = mdtraj.load(str(path_to_initial_pdb)).openmm_positions(0)

    dir_exists = pathlib.Path("./tmp_biasing_comp").exists()
    os.makedirs("./tmp_biasing_comp", exist_ok=True)
    path_to_traj = "./tmp_biasing_comp/traj_for_force.xtc"
    simulation.reporters.append(
        openmm_app.XTCReporter(path_to_traj, reportInterval=stride)
    )
    simulation.step(n_steps)

    traj = mdtraj.load(path_to_traj, top=path_to_initial_pdb)

    simulation = make_simulation_fn(parameters_for_md, pdb.topology)
    md_forces = _compute_regular_force(traj, simulation)

    simulation = make_simulation_fn(parameters_for_md, pdb.topology)
    bias_forces = _compute_biasing_force(
        traj, simulation, restrain_atom_list, positions_for_bias
    )

    k_value = _compute_k_value(
        jnp.array(md_forces),
        jnp.array(bias_forces),
        target_percentage,
        restrain_atom_list,
    ).mean()

    os.remove(path_to_traj)
    if not dir_exists:
        shutil.rmtree("./tmp_biasing_comp")

    return k_value


@partial(jax.vmap, in_axes=(0, 0, None, None))
def _compute_k_value(base_force, bias_force, target_percentage, restrain_atom_list):
    force1 = base_force[restrain_atom_list, :].flatten()
    force2 = bias_force[restrain_atom_list, :].flatten()

    rho = jnp.dot(force1, force2) / jnp.sum(force2**2)
    R = jnp.sum(force1**2) / jnp.sum(force2**2)

    a = 1.0 - target_percentage**2
    b = -2.0 * target_percentage**2 * rho
    c = -R * target_percentage**2

    return -b + jnp.sqrt(b**2 - 4.0 * a * c) / (2.0 * a)


def _compute_regular_force(trajectory: mdtraj.Trajectory, simulation):
    forces = np.zeros((trajectory.n_frames, trajectory.n_atoms, 3))
    for i in range(trajectory.n_frames):
        simulation.context.setPositions(trajectory.openmm_positions(i))
        forces[i] = np.array(
            simulation.context.getState(getForces=True).getForces(asNumpy=True)
        )
    return forces


def _compute_biasing_force(
    trajectory: mdtraj.Trajectory,
    simulation,
    restrain_atom_list,
    bias_positions,
):
    RMSD_value = openmm.RMSDForce(
        bias_positions,
        restrain_atom_list,
    )

    force_RMSD = openmm.CustomCVForce("0.5 * k * RMSD^2")
    force_RMSD.addGlobalParameter("k", 1.0)
    force_RMSD.addCollectiveVariable("RMSD", RMSD_value)
    simulation.system.addForce(force_RMSD)

    n_forces = len(simulation.system.getForces())
    for i in range(n_forces - 1):
        simulation.system.removeForce(0)

    simulation.context.reinitialize()
    forces = np.zeros((trajectory.n_frames, trajectory.n_atoms, 3))
    for i in range(trajectory.n_frames):
        simulation.context.setPositions(trajectory.openmm_positions(i))
        forces[i] = np.array(
            simulation.context.getState(getForces=True).getForces(asNumpy=True)
        )
    return forces


def _validate_base_state_file_path(base_state_file_path: str) -> str:
    # check if the path exists
    base_dir = os.path.dirname(base_state_file_path)
    if not os.path.exists(base_dir):
        os.makedirs(base_dir)

    if not base_state_file_path.endswith("_"):
        return f"{base_state_file_path}_it"
    else:
        return f"{base_state_file_path}it"


def _get_next_state_file_path(
    base_state_file_path: str,
    curr_state_file: str,
) -> str:
    # get the run counter from the last file
    last_counter = int(curr_state_file.split(base_state_file_path)[-1].split(".xml")[0])
    return f"{base_state_file_path}{last_counter + 1}.xml"


def _assert_is_valid_state_file(
    state_file: str,
    base_state_file_path: str,
) -> None:
    assert base_state_file_path in state_file, (
        "State file does not match base state file path. "
        "Please provide a valid state file."
    )

    counter = state_file.split(base_state_file_path)[-1].split(".xml")[0]
    try:
        int(counter)
    except ValueError:
        raise ValueError(
            f"State file should be formatted as base_state_file_path + <int>.xml. "
            f"Got {state_file} instead."
        )
    return


def _add_restraint_force_to_simulation(
    simulation: openmm_app.Simulation,
    positions: openmm_unit.Quantity,
    restrain_atom_list: list[int],
    bias_constant_in_kj_per_mol_angs: float,
) -> openmm_app.Simulation:
    RMSD_value = openmm.RMSDForce(
        positions,
        restrain_atom_list,
    )

    force_RMSD = openmm.CustomCVForce("0.5 * k * RMSD^2")
    force_RMSD.addGlobalParameter("k", bias_constant_in_kj_per_mol_angs)
    force_RMSD.addCollectiveVariable("RMSD", RMSD_value)
    simulation.system.addForce(force_RMSD)

    return simulation


def _remove_last_force_from_simulation(
    simulation: openmm_app.Simulation,
) -> openmm_app.Simulation:
    n_forces = len(simulation.system.getForces())
    simulation.system.removeForce(n_forces - 1)
    return simulation


def _default_make_sim_fn(parameters_for_md: dict, topology) -> openmm_app.Simulation:
    forcefield = _create_forcefield(parameters_for_md)
    integrator = _create_integrator(parameters_for_md)
    platform = _create_platform(parameters_for_md)
    system = _create_system(parameters_for_md, forcefield, topology)

    simulation = openmm_app.Simulation(
        topology,
        system,
        integrator,
        platform,
        parameters_for_md["properties"],
    )

    return simulation


def _create_forcefield(parameters_for_md: dict) -> openmm_app.ForceField:
    return openmm_app.ForceField(
        parameters_for_md["forcefield"], parameters_for_md["water_model"]
    )


def _create_integrator(parameters_for_md: dict) -> openmm.Integrator:
    return openmm.LangevinIntegrator(
        parameters_for_md["temperature"],
        parameters_for_md["friction"],
        parameters_for_md["timestep"],
    )


def _create_system(
    parameters_for_md: dict,
    forcefield: openmm_app.ForceField,
    topology: openmm_app.Topology,
) -> openmm.System:
    system = forcefield.createSystem(
        topology,
        nonbondedMethod=parameters_for_md["nonbondedMethod"],
        nonbondedCutoff=parameters_for_md["nonbondedCutoff"],
        constraints=parameters_for_md["constraints"],
    )

    return system


def _create_platform(parameters_for_md: dict) -> openmm.Platform:
    return openmm.Platform.getPlatformByName(parameters_for_md["platform"])


def _validate_and_set_params_for_md(
    parameters_for_md: dict,
) -> dict:
    default_md_params = _get_default_md_params()
    assert set(parameters_for_md.keys()).issubset(default_md_params)
    for key, value in default_md_params.items():
        if key not in parameters_for_md:
            parameters_for_md[key] = value

    return parameters_for_md



import cryojax.simulator as cxs
import jax.numpy as jnp
from cryojax.io import read_array_from_mrc, read_atoms_from_pdb

from cryojax_eo.utils import (
    Gaussian3D,
    ModelToVolumeAligner,
    fit_gmm_model_to_voxel_grid,
    make_gmm_model_from_atomic_model,
)


### GMM Fitting ####
def test_fit_gmm_to_density(sample_path_gmm_model, sample_path_mrc_file):
    gmm_positions = jnp.load(sample_path_gmm_model)["positions"]
    target_volume, voxel_size = read_array_from_mrc(
        sample_path_mrc_file, loads_grid_spacing=True
    )

    initial_gmm = Gaussian3D(
        positions=gmm_positions,
        amplitude=6.0,
        variance=1.0,
        shape=target_volume.shape,
        voxel_size=voxel_size,
        n_gaussians_per_bead=1,
    )
    fitted_gmm = fit_gmm_model_to_voxel_grid(initial_gmm, target_volume)

    loss_initial = jnp.linalg.norm(target_volume - initial_gmm.to_real_voxel_grid())

    render_fn = cxs.GaussianMixtureRenderFn(target_volume.shape, voxel_size)
    fitted_volume = render_fn(fitted_gmm)
    loss_fitted = jnp.linalg.norm(target_volume - fitted_volume)

    assert (
        loss_fitted < loss_initial
    ), "Fitted GMM should have lower loss than initial GMM"
    return


def test_make_gmm_model_from_atomic_model(sample_path_groel_pdb):
    gmm_model = make_gmm_model_from_atomic_model(
        sample_path_groel_pdb,
        box_size=32,
        voxel_size=4.0,
        fit_selection_string="name CA",
    )

    assert gmm_model.positions.ndim == 2, "GMM positions should be a 2D array"
    assert gmm_model.amplitudes.ndim == 2, "GMM amplitudes should be a 2D array"
    assert gmm_model.variances.ndim == 2, "GMM variances should be a 2D array"
    assert isinstance(
        gmm_model, cxs.GaussianMixtureVolume
    ), "GMM model should be a GaussianMixtureVolume"

    return


def test_volume_aligner(sample_path_groel_pdb):
    positions, _ = read_atoms_from_pdb(
        sample_path_groel_pdb, selection_string="name CA", center=True
    )

    render_fn = cxs.GaussianMixtureRenderFn((32, 32, 32), 4.0)
    real_voxel_grid = render_fn(cxs.GaussianMixtureVolume(positions, 1.0, 3.0))

    aligner = ModelToVolumeAligner(real_voxel_grid=real_voxel_grid, voxel_size=4.0)

    rotation = cxs.EulerAnglePose(phi_angle=1.0, theta_angle=3.0, psi_angle=3.0)
    rot_mtx = rotation.rotation.as_matrix()

    rotated_positions = positions @ rot_mtx

    _, solution = aligner.align(rotated_positions, 1.0, 3.0)

    cos_theta = (jnp.trace(rot_mtx @ solution.rotation_matrix) - 1) / 2.0

    assert jnp.isclose(cos_theta, 1.0), "Alignment to volume failed"
    return



from ._gmm_fitting import (
    Gaussian3D as Gaussian3D,
    fit_gmm_model_to_voxel_grid as fit_gmm_model_to_voxel_grid,
    make_gmm_model_from_atomic_model as make_gmm_model_from_atomic_model,
)
from ._rigid_body_align import ModelToVolumeAligner as ModelToVolumeAligner
from .rmsd_alignment import rigid_align_positions as rigid_align_positions



import cryojax.simulator as cxs
import jax.numpy as jnp
from equinox import Module
from jaxtyping import Array, Float


class DilatedMask(Module):
    volume: cxs.FourierVoxelGridVolume
    image_config: cxs.AbstractImageConfig

    def __init__(
        self,
        real_voxel_grid: Float[Array, "z y x"],
        image_config: cxs.AbstractImageConfig,
    ):
        self.image_config = image_config
        self.volume = cxs.FourierVoxelGridVolume.from_real_voxel_grid(real_voxel_grid)

    def project(self, pose: cxs.AbstractPose):
        mask2d = cxs.make_image_model(
            self.volume, self.image_config, pose.to_inverse_rotation()
        ).simulate()
        mask2d /= mask2d.max()
        mask2d = jnp.where(jnp.abs(mask2d) > 0.2, 1.0, 0.0)
        return mask2d



from functools import partial
from pathlib import Path
from typing import Annotated, Literal

import jax.numpy as jnp
import mdtraj
from pydantic import (
    AfterValidator,
    BaseModel,
    DirectoryPath,
    Field,
    FilePath,
    PositiveFloat,
    PositiveInt,
    field_serializer,
    field_validator,
    model_validator,
)

from .utils import _validate_file_with_type, _validate_files_with_type


class EnsOptMDConfigOptimizationConfig(BaseModel, extra="forbid"):
    n_steps: PositiveInt = Field(
        default=1, description="Number of steps for the optimization process."
    )
    step_size: PositiveFloat = Field(
        description="Step size in Angstroms for the optimization process."
    )
    n_batches_per_step: PositiveInt = Field(
        default=1,
        description="Number of batches to use for each optimization step. "
        + "Using more batches will provide a better estimate of the gradients, "
        + "but will also increase the computational cost. "
        + "If set to 1, the optimization will be performed using the entire dataset.",
    )
    batch_size: PositiveInt = Field(
        description="Batch size for SGD",
    )
    initial_weights: list[float] | None = Field(
        default=None,
        description="Initial weights for the models. "
        "If None, will be set to uniform distribution.",
    )
    estimates_pose: bool = Field(
        default=False,
        description="Whether to estimate the pose of the particles during optimization. "
        + "If True, the pose will be estimated using the current weights of the ensemble."
        + " If False, the pose will be estimated using uniform weights.",
    )

    @field_serializer("initial_weights")
    def serialize_initial_weights(self, v):
        if v is not None:
            v = jnp.array(v)
            v = v / jnp.sum(v)

        return v

    @field_validator("estimates_pose")
    @classmethod
    def validate_estimates_pose(cls, v):
        if v:
            raise Warning(
                "estimates_pose is set to True. This feature is still experimental, "
                + "and may slow down the optimization process."
            )
        return v


class EnsOptMDConfigProjector(BaseModel, extra="forbid"):
    n_steps: PositiveInt = Field(
        description="Number of steps for the MD sampler. Must be greater than 0."
    )
    bias_constant_in_kjpermol: PositiveFloat | list[PositiveFloat] = Field(
        description="Biasing constant for the projection step. "
        + "Can be a single value or a list of two values for linear scheduling."
    )
    platform: Literal["CPU", "CUDA", "OpenCL"] = Field(
        default="CPU",
        description="Platform to use for the MD sampler. "
        + "Must be 'CPU', 'CUDA', or 'OpenCL'.",
    )
    platform_properties: dict[str, str] | None = Field(
        default=None,
        description="Platform properties for OpenMM. "
        + "For CPU the default is {'Threads': '1'}"
        + " and for CUDA the default is {'DeviceIndex': '0'}.",
    )

    path_to_initial_states: str | list[FilePath] | None = Field(
        default=None,
        description="Path to the initial states. "
        + "If None, will be set to the path to the atomic models.",
    )

    @model_validator(mode="after")
    def validate_config(self):
        if self.platform == "CPU":
            if self.platform_properties is None:
                self.platform_properties = {"Threads": "1"}
        elif self.platform == "CUDA":
            if self.platform_properties is None:
                self.platform_properties = {"DeviceIndex": "0"}
        else:
            if self.platform_properties is None:
                raise ValueError(
                    "platform_properties must be provided for OpenCL platform"
                )
        return self

    @field_validator("path_to_initial_states")
    @classmethod
    def validate_path_to_initial_states(cls, v):
        if v is None:
            return v
        else:
            return _validate_files_with_type(v, file_types=[".xml"])


class EnsOptAlignConfig(BaseModel, extra="forbid"):
    path_to_prealigned_atomic_model: Annotated[
        str, AfterValidator(partial(_validate_file_with_type, file_type=".pdb"))
    ] = Field(
        description="Path to the reference model. "
        + "This model should be aligned to the cryo-EM particles, "
        + " and will be used for alignment of the walkers during optimization."
    )

    path_to_reference_volume: FilePath | None = Field(
        default=None,
        description="Path to the consensus volume. "
        + "Used for rigid body alignment of walkers. "
        + "If None, no alignment will be performed.",
    )

    downsample_box_size: PositiveInt = Field(
        default=32,
        description="Box size to downsample the volume for alignment. "
        + "The box size must be a positive integer.",
    )

    reference_volume_voxel_size: PositiveFloat | None = Field(
        default=None,
        description="Overrides the voxel size stored in the MRC header "
        + "of the consensus volume."
        + " If None, the voxel size from the MRC header will be used.",
    )

    @field_validator("path_to_prealigned_atomic_model")
    @classmethod
    def validate_path_to_prealigned_atomic_model(cls, v):
        return _validate_file_with_type(v, file_type=".pdb")

    @field_validator("path_to_reference_volume")
    @classmethod
    def validate_path_to_consensus_volume(cls, v):
        if v is not None:
            return _validate_file_with_type(v, file_type=".mrc")


class EnsOptDataConfig(BaseModel, extra="forbid"):
    path_to_starfile: FilePath = Field(
        description="Path to the starfile containing the particle information."
    )
    path_to_relion_project: DirectoryPath = Field(
        description="Path to the relion project directory."
    )
    loads_envelope: bool = Field(
        default=False, description="Whether to load the envelope from the starfile. "
    )
    path_to_volumetric_mask: FilePath | None = Field(
        default=None,
        description="Path to a volumetric mask. "
        + "For example: the dillated mask obtained from a homogeneous refinment job, "
        + "or a mask to focus on a specific region. "
        + "If None, no mask will be applied.",
    )
    data_sign: Literal["dark-on-light", "light-on-dark"] = Field(
        default="dark-on-light",
        description="Sign convention for the data. "
        + "'dark-on-light' means that the particles "
        + "are dark on a light background (default). "
        + "'light-on-dark' means that the particles "
        + "are light on a dark background.",
    )

    @field_validator("path_to_starfile")
    @classmethod
    def validate_path_to_starfile(cls, v):
        return _validate_file_with_type(v, file_type=".star")

    @field_validator("path_to_volumetric_mask")
    @classmethod
    def validate_path_to_volumetric_mask(cls, v):
        if v is not None:
            return _validate_file_with_type(v, file_type=".mrc")


class EnsOptMDConfig(BaseModel, extra="forbid"):
    # I/O
    path_to_atomic_models: str | list[FilePath] = Field(
        description="Path to the atomic models directory. "
        + "If a pattern is provided, all files matching the pattern will be used."
    )

    path_to_output: Path = Field(
        description="Path to the output directory. "
        + "If it does not exist, it will be created.",
    )
    atom_selection: str | FilePath = Field(
        default="all",
        description="Selection string for atom selection, "
        + "or a txt/npy file containing atom indices",
    )

    loads_b_factors: bool = Field(
        default=False,
        description="Whether to load the thermal b-factors from the PDB file. "
        + "Only used if the atomic model is in PDB format. "
        + "Otherwise it will be ignored."
        + "Also known as Debye-Waller factors.",
    )

    # Data
    data_params: dict = Field(
        description="Parameters for the experimental data. "
        + "This is a dictionary formatted by the `EnsOptDataConfig` class."
    )

    # Pipeline
    projector_params: dict = Field(
        description="Parameters for the physics-based ensemble projector. "
        + "This is a dictionary formatted by the `EnsOptMDConfigProjector` class."
    )
    likelihood_optimizer_params: dict = Field(
        description="Parameters for the ensemble optimizer. "
        + "This is a dictionary formatted by "
        + "the `EnsOptMDConfigOptimizationConfig` class."
    )

    alignment_params: dict = Field(
        description="Parameters for the alignment of the walkers. "
    )
    # Optimization
    n_steps: PositiveInt = Field(
        description="Number of steps of cryoJAX ensemble refinement to run."
    )

    # Miscellaneous
    rng_seed: int = Field(default=0, description="Random seed.")

    @model_validator(mode="after")
    def validate_config(self):
        n_atomic_models = len(self.path_to_atomic_models)
        if self.projector_params["path_to_initial_states"] is not None:
            n_initial_states = len(self.projector_params["path_to_initial_states"])
            assert n_atomic_models == n_initial_states, (
                f"Number of initial states {n_initial_states} "
                + f"does not match number of atomic models {n_atomic_models}."
            )

        if self.likelihood_optimizer_params["initial_weights"] is not None:
            n_initial_weights = len(self.likelihood_optimizer_params["initial_weights"])
            if n_atomic_models != n_initial_weights:
                raise Warning(
                    f"Number of initial weights {n_initial_weights} "
                    + f"does not match number of atomic models {n_atomic_models}."
                    + " Setting initial weights to uniform distribution."
                )
            self.likelihood_optimizer_params["initial_weights"] = jnp.asarray(
                [1.0 / n_atomic_models for _ in range(n_atomic_models)]
            )
        return self

    @field_validator("path_to_atomic_models")
    @classmethod
    def validate_path_to_atomic_models(cls, v):
        return _validate_files_with_type(v, file_types=[".pdb"])

    @field_validator("likelihood_optimizer_params")
    @classmethod
    def validate_ensemble_opt_config(cls, values):
        return dict(EnsOptMDConfigOptimizationConfig(**values).model_dump())

    @field_validator("projector_params")
    @classmethod
    def validate_md_sampler_config(cls, values):
        return dict(EnsOptMDConfigProjector(**values).model_dump())

    @field_validator("alignment_params")
    @classmethod
    def validate_aligner_config(cls, values):
        return dict(EnsOptAlignConfig(**values).model_dump())

    @field_validator("data_params")
    @classmethod
    def validate_data_config(cls, values):
        return dict(EnsOptDataConfig(**values).model_dump())

    @field_validator("atom_selection")
    @classmethod
    def validate_atom_selection(cls, values):
        suffix = Path(values).suffix
        if suffix in [".txt", ".npy"]:
            assert Path(values).exists(), f"Indices File: {values} does not exist."

        elif suffix not in [".txt", ".npy", ""]:
            raise ValueError("Invalid file type for atom selection.")

        else:
            try:
                mdtraj.Topology().select(values)
            except Exception as e:
                raise ValueError(f"Invalid atom selection string: {values}. Error: {e}")
        return values


### Keeping just in case we want to re-enable auto-incrementing output paths ###

# @field_validator("path_to_output")
# @classmethod
# def serialize_output_path(cls, v):
#     if not os.path.exists(v):
#         new_path = os.path.join(v, "Job001")

#     else:
#         # list all subdirectories
#         subdirs = [f for f in os.listdir(v) if os.path.isdir(os.path.join(v, f))]

#         # get the last job number
#         job_numbers = []
#         for subdir in subdirs:
#             if subdir.startswith("Job"):
#                 job_numbers.append(int(subdir[3:]))
#         job_numbers = sorted(job_numbers)
#         if len(job_numbers) == 0:
#             new_path = os.path.join(v, "Job001")
#         else:
#             new_path = os.path.join(v, f"Job{job_numbers[-1] + 1:03d}")

#     return new_path



"""
Weight and position optimizers for ensemble refinement.
"""

from typing import cast
from typing_extensions import override

import cryojax.simulator as cxs
import equinox as eqx
import jax
import jax.numpy as jnp
import jax_dataloader as jdl
from jaxopt import ProjectedGradient
from jaxopt.projection import projection_simplex
from jaxtyping import Array, Float, Int

from .._pose_search import HierarchicalSO3GridSearch
from . import ImagesToEnsembleLikelihoodFn
from ._utils import _estimate_poses
from .base_optimizer import AbstractEnsembleParameterOptimizer


class ProjGradDescWeightOptimizer(AbstractEnsembleParameterOptimizer):
    n_steps: Int
    ensemble_likelihood_fn: ImagesToEnsembleLikelihoodFn
    pose_search: HierarchicalSO3GridSearch | None

    def __init__(
        self,
        n_steps: Int,
        ensemble_likelihood_fn: ImagesToEnsembleLikelihoodFn,
        pose_search: HierarchicalSO3GridSearch | None = None,
    ):
        self.n_steps = n_steps
        self.ensemble_likelihood_fn = ensemble_likelihood_fn
        self.pose_search = pose_search

    @override
    def __call__(
        self,
        walkers: Float[Array, "n_walkers n_atoms 3"],
        weights: Float[Array, " n_walkers"],
        dataloader: jdl.DataLoader,
    ):
        """
        Optimize the weights of the walkers using projected gradient descent
        using all images.

        **Arguments:**
            walkers: The current positions of the walkers.
            weights: The current weights of the walkers.
            dataloader: The dataloader for the data.
            args: Additional arguments for the likelihood function.
            This should be a tuple with the following elements:
                - `amplitudes`: The Gaussian amplitudes for each atom.
                - `variances`: The Gaussian variances for each atom.
                - `noise_variance`: The noise variance for the data. If None, the
                    noise variance is marginalized.

        **Returns:**
            The optimized weights of the walkers.
        """
        likelihood_matrix = _compute_full_likelihood_matrix(
            walkers, dataloader, self.ensemble_likelihood_fn, self.pose_search
        )
        weights = _optimize_weights(weights, likelihood_matrix, self.n_steps)
        return weights


@eqx.filter_jit
def _optimize_weights(
    weights: Float[Array, " n_walkers"],
    likelihood_matrix: Float[Array, "n_images n_walkers"],
    n_steps: Int = 500,
) -> Float[Array, " n_walkers"]:
    pg = ProjectedGradient(
        fun=_compute_neg_log_likelihood,
        projection=projection_simplex,
        maxiter=n_steps,
    )
    return pg.run(weights, likelihood_matrix=likelihood_matrix).params


def _compute_full_likelihood_matrix(
    walkers: Float[Array, "n_walkers n_atoms 3"],
    dataloader: jdl.DataLoader,
    ensemble_likelihood_fn: ImagesToEnsembleLikelihoodFn,
    pose_search: HierarchicalSO3GridSearch | None,
) -> Array:
    """
    Compute the full likelihood matrix for the given walkers and dataloader.
    """

    shuffle = dataloader.dataloader.shuffle  # save the original shuffle state
    dataloader.dataloader.shuffle = False
    # Compute the likelihood matrix for each batch in the dataloader
    likelihood_matrix = []
    for batch in dataloader:
        if pose_search is None:
            poses_per_walker = jax.tree.map(
                lambda x: jnp.repeat(x[None, :], repeats=walkers.shape[0], axis=0),
                batch["particle_stack"]["parameters"]["pose"],
            )
            poses_per_walker = cast(cxs.EulerAnglePose, poses_per_walker)
        else:
            poses_per_walker = _estimate_poses(
                walkers,
                ensemble_likelihood_fn.image_to_walker_likelihood_fn.amplitudes,
                ensemble_likelihood_fn.image_to_walker_likelihood_fn.variances,
                batch["particle_stack"]["images"],
                batch["particle_stack"]["parameters"]["image_config"],
                batch["particle_stack"]["parameters"]["transfer_theory"],
                pose_search,
                n_walkers_in_parallel=ensemble_likelihood_fn.n_walkers_in_parallel,
                n_images_in_parallel=ensemble_likelihood_fn.n_images_in_parallel,
            )

        lklhood_matrix = ensemble_likelihood_fn.compute_log_likelihood_matrix(
            walkers,
            batch["particle_stack"]["images"],
            batch["particle_stack"]["parameters"]["image_config"],
            poses_per_walker,
            batch["particle_stack"]["parameters"]["transfer_theory"],
            batch["per_particle_args"],
        )
        likelihood_matrix.append(lklhood_matrix)

    # restore the original shuffle state
    dataloader.dataloader.shuffle = shuffle

    # Concatenate the likelihood matrices from all batches
    return jnp.concatenate(likelihood_matrix, axis=0)


def _compute_neg_log_likelihood(
    weights: Float[Array, " n_walkers"],
    likelihood_matrix: Float[Array, "n_images n_walkers"],
) -> Float:
    """
    Compute the negative log likelihood of the data given the weights and the
    likelihood matrix.

    **Arguments:**
        weights: The weights of the walkers.
        likelihood_matrix: The likelihood matrix for each image and walker.

    **Returns:**
        The negative log likelihood of the data given the weights and the
        likelihood matrix.
    """
    log_lklhood = jax.scipy.special.logsumexp(
        a=likelihood_matrix, b=weights[None, :], axis=1
    )
    return -jnp.mean(log_lklhood)



from typing import Any

import jax_dataloader as jdl
from cryospax import (
    RelionParticleDataset,
)
from jaxtyping import PRNGKeyArray

from cryojax_eo.typing import ParticleStackInfo


class CustomJaxDataset(jdl.Dataset):
    cryojax_dataset: RelionParticleDataset
    per_particle_args: Any | None

    def __init__(
        self,
        cryojax_dataset: RelionParticleDataset,
        per_particle_args: Any | None = None,
    ):
        self.cryojax_dataset = cryojax_dataset
        self.per_particle_args = per_particle_args

    def __getitem__(self, index) -> dict[str, ParticleStackInfo | Any]:
        if self.per_particle_args is None:
            per_particle_args = None
        else:
            per_particle_args = self.per_particle_args[index]

        data = {
            "particle_stack": self.cryojax_dataset[index],
            "per_particle_args": per_particle_args,
        }
        return data

    def __len__(self) -> int:
        return len(self.cryojax_dataset)


def create_dataloader(
    relion_stack_dataset: RelionParticleDataset,
    batch_size: int,
    shuffle: bool = False,
    drop_last: bool = False,
    *,
    per_particle_args: Any | None = None,
    jax_prng_key: PRNGKeyArray | None = None,
) -> jdl.DataLoader:
    """
    Create a Jax DataLoader for a `RelionParticleDataset`.

    **Arguments:**
        - relion_stack_dataset: A `RelionParticleDataset` object.
        - batch_size: The size of each batch.
        - shuffle: Whether to shuffle the dataset.
        - drop_last: Whether to drop the last batch if it is smaller than batch_size.
        - per_particle_args: Optional per-particle arguments to be passed to the dataset.
        - jax_prng_key: JAX PRNG key for shuffling. Required if shuffle is True.

    **Returns:**
        - A `jax_dataloader.DataLoader` instance. Where the dataset is a
          `cryoJAX` `RelionParticleStackDataset`.
    """

    if shuffle is True:
        if jax_prng_key is None:
            raise ValueError("jax_prng_key must be provided when shuffle is True.")
        dataloader = jdl.DataLoader(
            CustomJaxDataset(relion_stack_dataset, per_particle_args),
            backend="jax",
            batch_size=batch_size,
            shuffle=shuffle,
            drop_last=drop_last,
            generator=jax_prng_key,
        )
    else:
        dataloader = jdl.DataLoader(
            CustomJaxDataset(relion_stack_dataset, per_particle_args),
            backend="jax",
            batch_size=batch_size,
            shuffle=shuffle,
            drop_last=drop_last,
        )
    return dataloader



import logging
import os
import pathlib
from typing import Any
from typing_extensions import override

import jax
import jax.numpy as jnp
import mdtraj
import numpy as np
import optax
from jax_dataloader import DataLoader
from jaxtyping import Array, Float, Int, PRNGKeyArray
from mdtraj.formats import XTCTrajectoryFile
from tqdm import tqdm

from cryojax_eo.utils import ModelToVolumeAligner, rigid_align_positions

from .._likelihood_optimization import (
    IterativeEnsembleLikelihoodOptimizer,
    ProjGradDescWeightOptimizer,
)
from .._prior_projection.base_prior_projector import AbstractEnsemblePriorProjector
from .base_pipeline import AbstractEnsembleOptimizationPipeline


class EnsembleOptimizationPipeline(AbstractEnsembleOptimizationPipeline, strict=True):
    """
    Ensemble refinement pipeline using OpenMM for molecular dynamics simulation.
    """

    prior_projector: AbstractEnsemblePriorProjector
    likelihood_optimizer: IterativeEnsembleLikelihoodOptimizer
    n_steps: int
    prealigned_structure: mdtraj.Trajectory
    model_to_volume_aligner: ModelToVolumeAligner | None
    atom_indices_for_opt: Int[Array, " n_atoms_for_opt"]
    runs_postprocessing: bool

    def __init__(
        self,
        prior_projector: AbstractEnsemblePriorProjector,
        likelihood_optimizer: IterativeEnsembleLikelihoodOptimizer,
        n_steps: int,
        prealigned_structure: mdtraj.Trajectory,
        atom_indices_for_opt: Int[Array, " n_atoms_for_opt"],
        model_to_volume_aligner: ModelToVolumeAligner | None = None,
        *,
        runs_postprocessing: bool = True,
    ):
        self.prior_projector = prior_projector
        self.likelihood_optimizer = likelihood_optimizer
        self.n_steps = n_steps
        self.prealigned_structure = prealigned_structure
        self.model_to_volume_aligner = model_to_volume_aligner
        self.atom_indices_for_opt = atom_indices_for_opt
        self.runs_postprocessing = runs_postprocessing

    @override
    def run(
        self,
        key: PRNGKeyArray,
        initial_walkers: Float[Array, "n_walkers n_atoms 3"],
        initial_weights: Float[Array, " n_walkers"],
        dataloader: DataLoader,
        bias_constant_scheduler: optax.Schedule,
        *,
        output_directory: str | pathlib.Path,
        initial_state_for_projector: Any = None,
    ) -> tuple[
        Float[Array, "n_steps n_walkers n_atoms 3"],
        Float[Array, "n_steps n_walkers"],
    ]:
        logging.info("Initializing projector...")
        md_states = self.prior_projector.initialize(initial_state_for_projector)
        logging.info("Projector initialized.")

        ref_positions = (
            np.asarray(self.prealigned_structure.xyz[0]) * 10.0
        )  # Convert from nm to Angstroms

        walkers = initial_walkers.copy()
        weights = initial_weights.copy()

        if walkers.ndim == 2:
            walkers = jnp.expand_dims(walkers, axis=0)

        if weights.ndim == 0:
            weights = jnp.expand_dims(weights, axis=0)

        logging.info("Preparing writers for output...")
        writers = [
            XTCTrajectoryFile(os.path.join(output_directory, f"traj_walker_{i}.xtc"), "w")
            for i in range(walkers.shape[0])
        ]
        logging.info("Writers prepared.")

        logging.info("Aligning walkers to reference structure...")
        walkers = _align_walkers_to_reference(
            walkers, ref_positions, self.atom_indices_for_opt
        )
        logging.info("Walkers aligned.")

        progress_bar = tqdm(range(self.n_steps), desc="Optimization Progress")
        # make the tqdm progress bar show the current neg_log_likelihood at each step
        for i in progress_bar:
            logging.info(f"Starting optimization step {i + 1}/{self.n_steps}...")

            logging.info("   Likelihood Optimization: ")
            neg_log_likelihood, tmp_walkers, weights = self.likelihood_optimizer(
                walkers[:, self.atom_indices_for_opt, :],
                weights,
                dataloader,
            )
            logging.info(f"   Negative Log-Likelihood: {neg_log_likelihood}")
            progress_bar.set_description(
                f"Iter {i + 1}/{self.n_steps}, "
                f"Neg Log-Likelihood: {neg_log_likelihood:.4f}"
            )

            walkers = walkers.at[:, self.atom_indices_for_opt, :].set(tmp_walkers)
            walkers.block_until_ready()
            walkers = jax.device_get(walkers)
            logging.info("   Likelihood Optimization done.")

            logging.info("   Prior Projection: ")
            walkers, md_states = self.prior_projector(
                walkers, md_states, bias_constant_scheduler(i)
            )
            logging.info("   Prior Projection done.")

            logging.info("   Aligning walkers to reference structure...")
            walkers = _align_walkers_to_reference(
                walkers, ref_positions, self.atom_indices_for_opt
            )
            logging.info("   Walkers aligned.")

            if self.model_to_volume_aligner is not None:
                logging.info("   Aligning walkers to volume...")
                walkers = _align_walkers_to_volume(
                    walkers,
                    self.model_to_volume_aligner,
                    self.atom_indices_for_opt,
                    self.likelihood_optimizer.ensemble_likelihood_fn.image_to_walker_likelihood_fn.amplitudes,
                    self.likelihood_optimizer.ensemble_likelihood_fn.image_to_walker_likelihood_fn.variances,
                )
                logging.info("   Walkers aligned to volume.")

            logging.info("   Writing trajectory to files...")
            for j in range(walkers.shape[0]):
                writers[j].write(walkers[j] / 10.0)

        logging.info("Optimization complete.")

        for writer in writers:
            writer.close()

        for i, walker in enumerate(walkers):
            mdtraj.Trajectory(
                xyz=walker / 10.0,
                topology=self.prealigned_structure.topology,
            ).save_pdb(os.path.join(output_directory, f"final_walker_{i}.pdb"))

        if self.runs_postprocessing:
            logging.info("Running postprocessing...")
            weight_optimizer = ProjGradDescWeightOptimizer(
                n_steps=500,
                ensemble_likelihood_fn=self.likelihood_optimizer.ensemble_likelihood_fn,
                pose_search=self.likelihood_optimizer.pose_search,
            )
            walkers, weights = self.postprocess(
                walkers, weights, dataloader, weight_optimizer
            )
            logging.info("Postprocessing complete.")
        return walkers, weights

    def postprocess(
        self,
        walkers: Float[Array, "n_walkers n_atoms 3"],
        weights: Float[Array, " n_walkers"],
        dataloader: DataLoader,
        weight_optimizer: ProjGradDescWeightOptimizer,
    ):
        """
        Postprocess the walkers and weights.
        """
        # Project the weights
        weights = weight_optimizer(
            walkers[:, self.atom_indices_for_opt],
            weights,
            dataloader,
        )

        return walkers, weights


def _align_walkers_to_reference(
    walkers: Float[Array, "n_walkers n_atoms 3"],
    ref_positions: Float[Array, "n_atoms 3"],
    atom_indices: Int[Array, " n_atoms_for_opt"],
) -> Float[Array, "n_walkers n_atoms 3"]:
    """
    Align the walkers to the reference structure.
    """

    aligned_walkers = np.zeros_like(walkers)
    for i in range(walkers.shape[0]):
        _, rot_matrix, displacement = rigid_align_positions(
            walkers[i, atom_indices], ref_positions[atom_indices]
        )
        aligned_walkers[i] = walkers[i] @ rot_matrix.T + displacement

    return jnp.asarray(aligned_walkers)


def _align_walkers_to_volume(
    walkers: Float[Array, "n_walkers n_atoms 3"],
    model_to_volume_aligner: ModelToVolumeAligner,
    atom_indices: Int[Array, " n_atoms_for_opt"],
    amplitudes,
    variances,
) -> Float[Array, "n_walkers n_atoms 3"]:
    """
    Align the walkers to the volume using the ModelToVolumeAligner.
    """
    for i in range(walkers.shape[0]):
        atomic_positions = walkers[i, atom_indices, :]

        _, solution = model_to_volume_aligner.align(
            atomic_positions,
            amplitudes,
            variances,
        )
        aligned_positions = walkers[i] @ solution.rotation_matrix + solution.offset

        walkers = walkers.at[i].set(aligned_positions)
    return walkers



import os
import shutil

import mdtraj
import numpy as np

from cryojax_eo.ensemble_optimization import (
    EnsembleSteeredMDSimulator,
    SteeredMDSimulator,
)


def make_steered_md_simulator(path_to_pdb):
    model = mdtraj.load(path_to_pdb)
    atom_list = model.topology.select("not element H")

    return SteeredMDSimulator(
        path_to_pdb,
        n_steps=10,
        restrain_atom_list=atom_list,
        parameters_for_md={"platform": "CPU", "properties": {"Threads": "4"}},
        base_state_file_path=os.path.join(
            os.path.dirname(__file__), "outputs/md_states", "state_it"
        ),
    )


def test_steered_md_simulator(sample_path_to_pdb1):
    model = mdtraj.load(sample_path_to_pdb1)
    ref_positions = model.xyz[0] * 10.0

    simulator = make_steered_md_simulator(sample_path_to_pdb1)
    state = simulator.initialize()
    new_positions, state = simulator(ref_positions, state, 10.0)

    assert (
        ref_positions.shape == new_positions.shape
    ), "Positions after projection have the wrong shape"
    shutil.rmtree(os.path.join(os.path.dirname(__file__), "outputs/"))
    return


def test_ensemble_steered_md_simulator(sample_path_to_pdb1, sample_path_to_pdb2):
    model1 = mdtraj.load(sample_path_to_pdb1)
    model2 = mdtraj.load(sample_path_to_pdb2)
    ref_positions = np.stack([model1.xyz[0], model2.xyz[0]], axis=0) * 10.0

    simulator = EnsembleSteeredMDSimulator(
        [
            make_steered_md_simulator(sample_path_to_pdb1),
            make_steered_md_simulator(sample_path_to_pdb2),
        ]
    )
    state = simulator.initialize()
    new_positions, state = simulator(ref_positions, state, 10.0)

    assert (
        ref_positions.shape == new_positions.shape
    ), "Positions after projection have the wrong shape"
    shutil.rmtree(os.path.join(os.path.dirname(__file__), "outputs/"))
    return



"""
Weight and position optimizers for ensemble refinement.
"""

from typing import Any, cast
from typing_extensions import override

import cryojax.simulator as cxs
import equinox as eqx
import jax
import jax.numpy as jnp
import jax_dataloader as jdl
from jaxtyping import Array, Float

from .._pose_search import HierarchicalSO3GridSearch
from . import ImagesToEnsembleLikelihoodFn
from ._utils import _estimate_poses
from ._weights_optimizer import _optimize_weights
from .base_optimizer import AbstractEnsembleParameterOptimizer


class IterativeEnsembleLikelihoodOptimizer(AbstractEnsembleParameterOptimizer):
    step_size: Float
    n_steps: int
    n_batches_per_step: int
    ensemble_likelihood_fn: ImagesToEnsembleLikelihoodFn
    pose_search: HierarchicalSO3GridSearch | None

    def __init__(
        self,
        step_size: Float,
        n_steps: int,
        n_batches_per_step: int,
        ensemble_likelihood_fn: ImagesToEnsembleLikelihoodFn,
        pose_search: HierarchicalSO3GridSearch | None = None,
    ):
        assert step_size > 0, "Step size must be positive."
        assert n_steps > 0, "Number of steps must be positive."
        assert n_batches_per_step > 0, "Number of batches per step must be positive."

        self.step_size = step_size
        self.n_steps = n_steps
        self.n_batches_per_step = n_batches_per_step
        self.ensemble_likelihood_fn = ensemble_likelihood_fn
        self.pose_search = pose_search

    @override
    def __call__(
        self,
        walkers: Float[Array, "n_walkers n_atoms 3"],
        weights: Float[Array, " n_walkers"],
        dataloader: jdl.DataLoader,
    ) -> tuple[float, Float[Array, "n_walkers n_atoms 3"], Float[Array, " n_walkers"]]:
        loss = 0.0
        for _ in range(self.n_steps):
            gradients = jnp.zeros_like(walkers)
            weights = jnp.ones_like(weights) / weights.shape[0]
            loss = 0.0
            for _ in range(self.n_batches_per_step):
                batch = next(iter(dataloader))
                if self.pose_search is None:
                    poses_per_walker = jax.tree.map(
                        lambda x: jnp.repeat(
                            x[None, :], repeats=walkers.shape[0], axis=0
                        ),
                        batch["particle_stack"]["parameters"]["pose"],
                    )
                    poses_per_walker = cast(cxs.EulerAnglePose, poses_per_walker)
                else:
                    poses_per_walker = _estimate_poses_per_walker(
                        walkers,
                        batch,
                        self.ensemble_likelihood_fn,
                        self.pose_search,
                    )

                tmp_loss, tmp_grads, tmp_weights = _compute_ensemble_gradients(
                    walkers,
                    weights,
                    batch["particle_stack"]["images"],
                    batch["particle_stack"]["parameters"]["image_config"],
                    poses_per_walker,
                    batch["particle_stack"]["parameters"]["transfer_theory"],
                    batch["per_particle_args"],
                    self.ensemble_likelihood_fn,
                )
                gradients += tmp_grads
                weights += tmp_weights
                loss += tmp_loss

            gradients /= self.n_batches_per_step
            weights /= self.n_batches_per_step
            loss /= self.n_batches_per_step

            norms = jnp.linalg.norm(gradients, axis=(2), keepdims=True)
            norms = jnp.where(norms < 1e-12, 1.0, norms)

            # norms = (
            #     jnp.linalg.norm(gradients, axis=(1, 2), keepdims=True)
            #     / gradients.shape[1]
            # )
            gradients /= norms

            walkers = walkers - self.step_size * gradients
        return loss, walkers, weights


def _estimate_poses_per_walker(
    walkers: Float[Array, "n_walkers n_atoms 3"],
    batch: Any,
    ensemble_likelihood_fn: ImagesToEnsembleLikelihoodFn,
    pose_search: HierarchicalSO3GridSearch,
):
    return _estimate_poses(
        walkers,
        ensemble_likelihood_fn.image_to_walker_likelihood_fn.amplitudes,
        ensemble_likelihood_fn.image_to_walker_likelihood_fn.variances,
        batch["particle_stack"]["images"],
        batch["particle_stack"]["parameters"]["image_config"],
        batch["particle_stack"]["parameters"]["transfer_theory"],
        pose_search,
        n_walkers_in_parallel=ensemble_likelihood_fn.n_walkers_in_parallel,
        n_images_in_parallel=ensemble_likelihood_fn.n_images_in_parallel,
    )


@eqx.filter_jit
def _compute_ensemble_gradients(
    walkers: Float[Array, "n_walkers n_atoms 3"],
    weights: Float[Array, " n_walkers"],
    images: Float[Array, "n_images y x"],
    image_config: cxs.BasicImageConfig,
    poses_per_walker: cxs.AbstractPose,
    transfer_theories: cxs.ContrastTransferTheory,
    per_particle_args: Any,
    ensemble_likelihood_fn: ImagesToEnsembleLikelihoodFn,
) -> tuple[
    float,
    Float[Array, "n_walkers n_atoms 3"],
    Float[Array, " n_walkers"],
]:
    """
    Optimize the walkers and weights of the ensemble. First, the weights
    are optimized through projected gradient descent, and then the walkers
    are optimized with steepest descent.

    **Arguments:**
        walkers: The current positions of the walkers.
        weights: The current weights of the walkers.
        images: The images to optimize against.
        image_config: Configuration for the images.
        poses_per_walker: The poses associated with each walker.
        transfer_theories: The contrast transfer theories.
        per_particle_args: Additional arguments for each particle.
        ensemble_likelihood_fn: The likelihood function for the ensemble.

    **Returns:**
        The optimized walkers and weights of the ensemble.
    """

    def _loss_fn(walkers, weights):
        log_likelihood_matrix = ensemble_likelihood_fn.compute_log_likelihood_matrix(
            walkers,
            images,
            image_config,
            poses_per_walker,
            transfer_theories,
            per_particle_args,
        )
        weights = jax.nn.softmax(_optimize_weights(weights, log_likelihood_matrix))
        log_lklhood = jax.scipy.special.logsumexp(
            a=log_likelihood_matrix, b=weights[None, :], axis=1
        )
        return -jnp.mean(log_lklhood), weights

    (loss, weights), grads = jax.value_and_grad(_loss_fn, argnums=0, has_aux=True)(
        walkers,
        weights,
    )
    return loss, grads, weights



from ._datasets_and_dataloaders import (
    create_dataloader as create_dataloader,
)
from ._relion_data_simulation import (
    make_relion_parameter_file as make_relion_parameter_file,
    simulate_relion_dataset as simulate_relion_dataset,
)



import warnings
from pathlib import Path
from typing import Literal

import jax.numpy as jnp
from pydantic import (
    BaseModel,
    Field,
    FilePath,
    PositiveFloat,
    PositiveInt,
    field_serializer,
    field_validator,
)

from .utils import _validate_files_with_type


class DatasetSimulatorConfigAtomicModels(BaseModel, extra="forbid"):
    """
    Parameter for loading the atomic models parameters used
    in the data generation pipeline.
    """

    path_to_atomic_models: str | list[FilePath] = Field(
        description="Path to the atomic models directory. "
        + "If a pattern is provided, all files matching the pattern will be used."
    )

    atomic_models_probabilities: PositiveFloat | list[PositiveFloat] = Field(
        description="Probabilstic weights for each model. Will be normalized to sum to 1."
    )

    loads_b_factors: bool = Field(
        default=False,
        description="Whether to load the B-factors from the PDB file. "
        + "Only used if the atomic model is in PDB format. "
        + "Otherwise it will be ignored.",
    )

    atom_selection: str = Field(
        default="all",
        description="Selection string for the atoms to use. "
        + "Only used if the atomic model is in PDB format. "
        + "Otherwise it will be ignored.",
    )

    @field_serializer("atomic_models_probabilities")
    def serialize_atomic_model_probabilities(self, v):
        if isinstance(v, int):
            v = [v]
        v = jnp.array(v)
        return v / jnp.sum(v)

    @field_serializer("path_to_atomic_models")
    def serialize_path_to_atomic_models(self, v):
        return _validate_files_with_type(v, file_types=[".pdb", ".npz"])


class DatasetSimulatorConfig(BaseModel, extra="forbid"):
    """
    Parameters for the data generation pipeline.

    If an item can be either a list or a single value, the list will be used as the range for random data generation. For example, if `offset_x_in_angstroms` is defined as `[0, 10]`, the offset in the x direction will be randomly generated between 0 and 10 for each image. If a single value is provided, the same value will be used for all images.

    """  # noqa

    # Experiment setup
    number_of_images: PositiveInt = Field(description="Number of images to generate.")

    data_sign: Literal["dark-on-light", "light-on-dark"] = Field(
        default="dark-on-light",
        description="Sign convention for the data. "
        + "'dark-on-light' means that the particles "
        + "are dark on a light background (default). "
        + "'light-on-dark' means that the particles "
        + "are light on a dark background.",
    )
    # Instrument
    pixel_size: PositiveFloat = Field(description="Pixel size in Angstroms.")
    box_size: PositiveInt = Field(description="Size of the simulation box in pixels.")
    pad_scale: PositiveInt = Field(
        1, description="Factor to scale the box size for padding."
    )
    voltage_in_kilovolts: PositiveFloat = Field(
        300.0, description="Voltage in kilovolts."
    )

    # Pose
    offset_x_in_angstroms: float | list[float] = Field(
        0.0, description="Offset in x direction in Angstroms."
    )
    offset_y_in_angstroms: float | list[float] = Field(
        0.0, description="Offset in y direction in Angstroms."
    )

    # Transfer Theory
    defocus_in_angstroms: PositiveFloat | list[PositiveFloat] = Field(
        0.0, description="Defocus in Angstroms."
    )
    astigmatism_in_angstroms: float | list[float] = Field(
        0.0, description="Astigmatism in Angstroms."
    )
    astigmatism_angle_in_degrees: float | list[float] = Field(
        0.0, description="Astigmatism angle in degrees."
    )
    phase_shift: float | list[float] = Field(0.0, description="Phase shift in radians.")
    amplitude_contrast_ratio: PositiveFloat = Field(
        0.1, description="Amplitude contrast ratio."
    )
    spherical_aberration_in_mm: PositiveFloat = Field(
        2.7, description="Microscope spherical aberration in mm."
    )
    ctf_scale_factor: PositiveFloat = Field(1.0, description="CTF scale factor.")
    envelope_b_factor: float | list[float] = Field(
        0.0, description="Envelope B-factor in Angstroms^2."
    )

    # Noise and randomness
    noise_snr: PositiveFloat | list[PositiveFloat] = Field(
        description="Signal to noise ratio."
    )
    mask_radius: PositiveFloat | None = Field(
        default=None,
        description="Radius for a circular cryojax Mask."
        + " This is used to compute the variance of the signal, "
        + "and then define the noise variance through the SNR. "
        + "If None, will be set to box_size // 3.",
    )
    mask_rolloff_width: PositiveFloat = Field(
        default=0.0, description="Width of the rolloff for the mask. "
    )

    rng_seed: int = Field(0, description="Seed for random number generation.")

    # Atomic modelss
    atomic_models_params: dict = Field(
        description="Parameters for the atomic models. This is a dictionary "
        + "formatted by the `DatasetSimulatorConfigAtomicModels` class."
    )

    # I/O
    path_to_relion_project: Path = Field(
        description="Path to the RELION project directory."
    )
    path_to_starfile: Path = Field(description="Path to the RELION star file.")
    images_per_file: PositiveInt = Field(description="Images per .mrcs.")
    batch_size_for_generation: PositiveInt = Field(
        default=1,
        description="Batch size for the data generation. "
        + "This is used to generate the data in batches.",
    )
    overwrite: bool = Field(False, description="Overwrite existing files if True.")

    @field_validator("atomic_models_params")
    @classmethod
    def validate_atomic_models_params(cls, v):
        return dict(DatasetSimulatorConfigAtomicModels(**v).model_dump())

    @field_serializer("offset_x_in_angstroms")
    def serialize_offset_x_in_angstroms(self, v):
        if isinstance(v, float):
            v = jnp.array([v, v])
        else:
            v = jnp.array(v)
        return v

    @field_serializer("offset_y_in_angstroms")
    def serialize_offset_y_in_angstroms(self, v):
        if isinstance(v, float):
            v = jnp.array([v, v])
        else:
            v = jnp.array(v)
        return v

    @field_serializer("defocus_in_angstroms")
    def serialize_defocus_in_angstroms(self, v):
        if isinstance(v, float):
            v = jnp.array([v, v])
        else:
            v = jnp.array(v)
        return v

    @field_serializer("astigmatism_in_angstroms")
    def serialize_astigmatism_in_angstroms(self, v):
        if isinstance(v, float):
            v = jnp.array([v, v])
        else:
            v = jnp.array(v)
        return v

    @field_serializer("astigmatism_angle_in_degrees")
    def serialize_astigmatism_angle(self, v):
        if isinstance(v, float):
            v = jnp.array([v, v])
        else:
            v = jnp.array(v)
        return v

    @field_serializer("phase_shift")
    def serialize_phase_shift(self, v):
        if isinstance(v, float):
            v = jnp.array([v, v])
        else:
            v = jnp.array(v)
        return v

    @field_serializer("envelope_b_factor")
    def serialize_envelope_b_factor(self, v):
        if isinstance(v, float):
            v = jnp.array([v, v])
        else:
            v = jnp.array(v)
        return v

    @field_serializer("noise_snr")
    def serialize_noise_snr(self, v):
        if isinstance(v, float):
            v = jnp.array([v, v])
        else:
            v = jnp.array(v)
        return v

    @field_serializer("mask_radius")
    def serialize_noise_radius_mask(self, v):
        if v is None:
            v = self.box_size // 3

        elif v > self.box_size:
            warnings.warn(
                "Noise radius mask is greater than box size. Setting to box size."
            )
            v = self.box_size

        return v



"""
Geometry related codes
The SO(3) grid part codes are adpated from deeprefine
(https://github.com/minhuanli/deeprefine/blob/master/deeprefine/geometry.py)
based on Hopf Fibration
All jax-ified functions have a `_jnp` suffix, other functions are numpy based
"""

from functools import partial

import jax
import jax.numpy as jnp
import numpy as np
from jaxtyping import Array, Float, Int

from .healjax import pix2ang


def construct_SO3_jnp(v1, v2):
    """
    Construct a continuous representation of SO(3) rotation with two 3D vectors
    https://arxiv.org/abs/1812.07035
    Parameters
    ----------
    v1, v2: 3D arrays
        Real-valued array in 3D space
    Returns
    -------
    R: A 3x3 SO(3) rotation matrix
    """
    e1 = v1 / jnp.linalg.norm(v1)
    u2 = v2 - e1 * jnp.tensordot(e1, v2, axes=1)
    e2 = u2 / jnp.linalg.norm(u2)
    e3 = jnp.cross(e1, e2)
    R = jnp.stack((e1, e2, e3)).T
    return R


def decompose_SO3(R, a=1, b=1, c=1):
    """
    Decompose the rotation matrix into the two vector representation
    This decomposition is not unique, so a, b, c can be set as
    arbitrary constants you like
    c != 0
    Parameters
    ----------
    R: 3x3 array
        Real-valued rotation matrix
    a, b, c: scalar constants
        Arbitrary constants for the decomposition (c must be nonzero)
    Returns
    -------
    v1, v2: Two real-valued 3D arrays, as the continuous
    representation of the rotation matrix
    """
    assert c != 0, "Give a nonzero c!"
    v1 = a * R[:, 0]
    v2 = b * R[:, 0] + c * R[:, 1]

    return v1, v2


# Quaternion representation of SO(3) and Hopf Fibration grid
def grid_s1(resol=1):
    Npix = 6 * 2**resol
    dt = 2 * np.pi / Npix
    grid = jnp.arange(Npix) * dt + dt / 2
    return grid


def grid_s2(resol=1):
    Nside = 2**resol
    Npix = 12 * Nside * Nside
    theta, phi = pix2ang(Nside, np.arange(Npix))
    return theta, phi


@partial(jax.vmap, in_axes=(0, 0, None))
@partial(jax.vmap, in_axes=(None, None, 0))
def _hopf_to_quat(theta, phi, psi):
    """
    Hopf coordinates to quaternions
    theta: [0,pi]
    phi: [0, 2pi)
    psi: [0, 2pi)
    already normalized
    """
    ct = jnp.cos(theta / 2)
    st = jnp.sin(theta / 2)
    quat = jnp.array(
        [
            ct * jnp.cos(psi / 2),
            ct * jnp.sin(psi / 2),
            st * jnp.cos(phi + psi / 2),
            st * jnp.sin(phi + psi / 2),
        ]
    )
    return quat.T.astype(np.float32)


def hopf_to_quat(theta, phi, psi):
    return _hopf_to_quat(theta, phi, psi).reshape(-1, 4)


def grid_SO3(resol):
    theta, phi = grid_s2(resol)
    psi = grid_s1(resol)
    quat = hopf_to_quat(theta, phi, psi).reshape(-1, 4)
    return quat


def quat_distance(q1, q2):
    """
    q1: [n1, 4]
    q2: [n2, 4]

    Return:
        [n1, n2]
    """
    q1 = q1 / np.linalg.norm(q1, ord=2, axis=-1, keepdims=True)
    q2 = q2 / np.linalg.norm(q2, ord=2, axis=-1, keepdims=True)
    args = np.abs(np.sum(q1[:, None, :] * q2[None, ...], axis=-1))
    return 2.0 * np.arccos(args)


def mat_distance(R1, R2):
    """
    R1: [a, 3, 3]
    R2: [b, 3, 3]

    Return:
        [a, b]
    """
    R = np.einsum("axy,bzy->abxz", R1, R2)
    args = (np.einsum("abii", R) - 1) / 2.0
    return np.arccos(args)


def quaternions_to_SO3_jnp(q):
    """
    Normalizes q and maps to group matrix.
    https://en.wikipedia.org/wiki/Quaternions_and_spatial_rotation#Quaternion-derived_rotation_matrix
    https://danceswithcode.net/engineeringnotes/quaternions/quaternions.html

    Parameters
    ----------
    q: JAX array of shape (..., 4)
        Quaternion(s) in format [w, x, y, z] where w is real part

    Returns
    -------
    R: JAX array of shape (..., 3, 3)
        SO(3) rotation matrix/matrices
    """
    # Normalize quaternion
    q = q / jnp.linalg.norm(q, axis=-1, keepdims=True)
    r, i, j, k = q[..., 0], q[..., 1], q[..., 2], q[..., 3]

    # Stack rotation matrix elements
    rotation_elements = jnp.stack(
        [
            1 - 2 * j * j - 2 * k * k,
            2 * (i * j - r * k),
            2 * (i * k + r * j),
            2 * (i * j + r * k),
            1 - 2 * i * i - 2 * k * k,
            2 * (j * k - r * i),
            2 * (i * k - r * j),
            2 * (j * k + r * i),
            1 - 2 * i * i - 2 * j * j,
        ],
        axis=-1,
    )

    # Reshape to 3x3 matrices
    return rotation_elements.reshape(*q.shape[:-1], 3, 3)


def SO3_to_quaternions_jnp(mats, key=None):
    """
    Convert SO(3) rotation matrices to quaternions.

    Parameters
    ----------
    mats: JAX array of shape (n, 3, 3)
        Rotation matrices
    key: JAX random key (optional)
        Random key for reproducible random vector generation
        If None, uses a fixed seed

    Returns
    -------
    qs: JAX array of shape (n, 4)
        Quaternions in format [w, x, y, z]
    """
    if key is None:
        key = jax.random.PRNGKey(42)  # Fixed seed for reproducibility

    # Compute eigenvalues and eigenvectors
    w, v = jnp.linalg.eig(mats)

    # Find indices where eigenvalue is close to 1 (rotation axis)
    # For real rotation matrices, one eigenvalue should be 1
    closest_to_one = jnp.argmin(jnp.abs(w - 1.0), axis=-1)
    batch_indices = jnp.arange(mats.shape[0])
    us = jnp.real(v[batch_indices, :, closest_to_one]).T

    # Generate random vector and compute orthogonal vector
    # take a random vector v orthogonal to u, determine sign of sin(theta/2)
    # https://math.stackexchange.com/questions/893984/conversion-of-rotation-matrix-to-quaternion
    v_rand = jax.random.normal(key, (3,))
    vs = jnp.cross(us, v_rand[None, :])
    vs = vs / jnp.linalg.norm(vs, axis=-1, keepdims=True)  # Normalize

    # Determine sign of sin(theta/2)
    cross_product = jnp.cross(vs, jnp.einsum("nij,nj->ni", mats, vs))
    signs = jnp.sign(jnp.einsum("ni,ni->n", cross_product, us))

    # Compute rotation angles
    traces = jnp.trace(mats, axis1=-2, axis2=-1)
    args = (traces - 1) / 2
    # Clamp to avoid numerical issues with arccos
    args = jnp.clip(args, -1.0, 1.0)
    thetas = jnp.arccos(args)

    # Compute quaternion components
    cos_halftheta = jnp.cos(thetas / 2.0)
    sin_halftheta = jnp.sin(thetas / 2.0) * signs

    # Construct quaternions [w, x, y, z]
    qs = jnp.concatenate([cos_halftheta[:, None], sin_halftheta[:, None] * us], axis=1)

    # Normalize quaternions
    qs = qs / jnp.linalg.norm(qs, axis=-1, keepdims=True)

    return qs


# Neighbors on the Hopf grid
def get_s1_neighbor(mini, curr_res):
    """
    Return the 2 nearest neighbors on S1 at the next resolution level
    """
    Npix = 6 * 2 ** (curr_res + 1)
    dt = 2 * np.pi / Npix
    # return np.array([2*mini, 2*mini+1])*dt + dt/2
    # the fiber bundle grid on SO3 is weird
    # the next resolution level's nearest neighbors in SO3 are not
    # necessarily the nearest neighbor grid points in S1
    # include the 13 neighbors for now... eventually learn/memorize the mapping
    ind = jnp.arange(0, 4, 1) + (2 * mini - 1)
    ind = jax.lax.cond(ind[0] < 0, lambda x: x.at[0].set(x[0] + Npix), lambda x: x, ind)

    return ind * dt + dt / 2, ind


def get_s2_neighbor(mini, curr_res):
    """
    Return the 4 nearest neighbors on S2 at the next resolution level
    """
    Nside = 2 ** (curr_res + 1)
    ind = jnp.arange(4) + 4 * mini
    return pix2ang(Nside, ind), ind


def get_base_ind(
    ind: Int[Array, " n_indices"], base_resol: Float = 1.0
) -> tuple[Int[Array, " n_indices"], Int[Array, " n_indices"]]:
    """
    Return the corresponding S2 and S1 grid index for an index on the base SO3 grid
    """
    psii = ind % (6 * 2**base_resol)
    thetai = ind // (6 * 2**base_resol)
    return thetai, psii


@partial(jax.vmap, in_axes=(0, 0, 0, None))
def get_neighbor_SO3(quat, s2i, s1i, curr_res):
    """
    Return the 8 nearest neighbors on SO3 at the next resolution level
    """
    (theta, phi), s2_nexti = get_s2_neighbor(s2i, curr_res)
    psi, s1_nexti = get_s1_neighbor(s1i, curr_res)
    quat_n = hopf_to_quat(theta, phi, psi)
    ind = jnp.array(
        [
            jnp.repeat(s2_nexti, len(psi), total_repeat_length=16),
            jnp.tile(s1_nexti, len(theta)),
        ]
    )
    ind = ind.T
    # find the 8 nearest neighbors of 16 possible points
    # need to check distance from both +q and -q
    dists = jnp.minimum(
        jnp.sum((quat_n - quat) ** 2, axis=1), jnp.sum((quat_n + quat) ** 2, axis=1)
    )
    _, best_8_indices = jax.lax.top_k(-dists, k=8)

    return quat_n[best_8_indices], ind[best_8_indices]


# Loss based neighbor search
def getbestneighbors_base_SO3(loss, base_quats, N=10, base_resol=1):
    _, bestN_index = jax.lax.top_k(-loss, k=N)
    best_quats = base_quats[bestN_index]
    s2_index, s1_index = get_base_ind(bestN_index, base_resol)
    allnb_quats, allnb_s2s1 = get_neighbor_SO3(best_quats, s2_index, s1_index, base_resol)
    return allnb_quats.reshape(-1, 4), allnb_s2s1.reshape(-1, 2)


def getbestneighbors_next_SO3(loss, quats, s2s1_arr, curr_res=2, N=50):
    _, bestN_index = jax.lax.top_k(-loss, k=N)
    best_quats = quats[bestN_index]

    s2_index = s2s1_arr[bestN_index, 0].astype(int)
    s1_index = s2s1_arr[bestN_index, 1].astype(int)
    allnb_quats, allnb_s2s1 = get_neighbor_SO3(best_quats, s2_index, s1_index, curr_res)
    return allnb_quats.reshape(-1, 4), allnb_s2s1.reshape(-1, 2)



#!/usr/bin/env python3
import argparse
import datetime
import logging
import os
import sys

import yaml

from ..dataset import simulate_relion_dataset
from ..internal import DatasetSimulatorConfig


def add_args(parser):
    parser.add_argument(
        "--config", type=str, default=None, help="Path to the config (yaml) file"
    )
    return parser


def mkbasedir(out):
    if not os.path.exists(out):
        try:
            os.makedirs(out)
        except (FileExistsError, PermissionError):
            raise ValueError("Output path does not exist and cannot be created.")
    return


def warnexists(out):
    if os.path.exists(out):
        Warning(f"Warning: {out} already exists. Overwriting.")


def main(args):
    with open(args.config) as f:
        config_dict = yaml.safe_load(f)
        config = DatasetSimulatorConfig(**config_dict)

    project_path = config.path_to_relion_project
    warnexists(project_path)
    mkbasedir(project_path)
    print(f"A copy of the config file and the log will be written to {project_path}")
    sys.stdout.flush()

    # make copy of config to output_path

    logger = logging.getLogger()
    logger_fname = datetime.datetime.now().strftime("%Y-%m-%d-%H")
    logger_fname = os.path.join(project_path, logger_fname + ".log")
    fhandler = logging.FileHandler(filename=logger_fname, mode="a")
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    fhandler.setFormatter(formatter)
    logger.addHandler(fhandler)
    logger.setLevel(logging.INFO)

    config_fname = os.path.basename(args.config)
    os.system(f"cp {args.config} {os.path.join(project_path, config_fname)}")

    logging.info(
        f"A copy of the used config file has been written"
        f" to {os.path.join(project_path, config_fname)}"
    )

    logging.info("Simulating particle stack...")
    simulate_relion_dataset(config)
    logging.info("Simulation complete.")
    logging.info(f"Output written to {project_path}")

    return


def main_cli():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=yaml.dump(DatasetSimulatorConfig.model_json_schema(), indent=4),
    )
    main(add_args(parser).parse_args())


if __name__ == "__main__":
    main_cli()



import cryojax.simulator as cxs
import jax
import jax.numpy as jnp
from cryojax.ndimage import CircularCosineMask
from jaxtyping import Array, Float, Int, PRNGKeyArray

from cryojax_eo.typing import ParticleParameterInfo


def _select_potential(volumes, idx):
    funcs = [lambda i=i: volumes[i] for i in range(len(volumes))]
    return jax.lax.switch(idx, funcs)


def simulate_image_with_white_gaussian_noise(
    particle_parameters: ParticleParameterInfo,
    constant_args: tuple[
        tuple[cxs.AbstractVolumeRepresentation],
        CircularCosineMask,
        float,
    ],
    per_particle_args: tuple[PRNGKeyArray, Int, Float],
) -> Float[
    Array,
    "{relion_particle_stack.config.y_dim} {relion_particle_stack.config.x_dim}",  # noqa
]:
    """
    Simulates an image given the particle parameters, volume,
    and noise variance. The noise is White Gaussian noise.

    **Arguments:**
        - `particle_parameters`: The particle parameters.
        - `constant_args`: A tuple containing: a tuple of volume parametrizations,
           a mask defining the signal region, and the data_sign.
           A data_sign of 1.0 means dark-on-light, -1.0 means light-on-dark.

        - `per_particle_args`: A containing a random jax key,
            the potential_idx to use, and the signal-to-noise ratio
    **Returns:**
        The simulated image.

    """
    key_noise, potential_idx, snr = per_particle_args
    volumes, mask, data_sign = constant_args
    volume = _select_potential(volumes, potential_idx)

    image_model = cxs.make_image_model(
        volume,
        particle_parameters["image_config"],
        particle_parameters["pose"],
        particle_parameters["transfer_theory"],
        signal_region=(mask.get() == 1),
        normalizes_signal=True,
    )

    distribution = cxs.GaussianWhiteNoiseModel(
        image_model,
        variance=1.0,
        signal_scale_factor=jnp.sqrt(snr),
    )
    return data_sign * distribution.sample(key_noise)



import cryojax.simulator as cxs
import equinox as eqx
from cryojax.jax_util import filter_bmap
from jaxtyping import Array, Float

from .._pose_search import HierarchicalSO3GridSearch


def _estimate_poses(
    walkers: Float[Array, "n_walkers n_atoms 3"],
    variances: Float[Array, "n_atoms n_gaussians"],
    amplitudes: Float[Array, "n_atoms n_gaussians"],
    images: Float[Array, "n_images y x"],
    image_config: cxs.BasicImageConfig,
    transfer_theories: cxs.ContrastTransferTheory,
    pose_search: HierarchicalSO3GridSearch,
    n_walkers_in_parallel: int,
    n_images_in_parallel: int,
) -> cxs.QuaternionPose:
    """
    Estimate the best poses for each walker and image using the pose search.

    **Arguments:**
    - walkers:
        The current positions of the walkers as Gaussian mixture volumes.
    - variances:
        The variances of the Gaussian mixtures.
    - amplitudes:
        The amplitudes of the Gaussian mixtures.
    - images:
        The images to estimate the poses for.
    - image_config:
        The configuration of the images.
    - transfer_theories:
        The transfer theories for the images.
    - pose_search:
        The pose search object.

    **Returns:**
        The estimated poses for each walker and image.
    """
    return filter_bmap(
        lambda x: _estimate_poses_for_walker(
            x,
            amplitudes,
            variances,
            images,
            image_config,
            transfer_theories,
            pose_search,
            n_images_in_parallel,
        ),
        xs=walkers,
        batch_size=n_walkers_in_parallel,
    )


@eqx.filter_vmap(in_axes=(0, None, None, None, None, None, None, None))
def _estimate_poses_for_walker(
    walker: Float[Array, "n_atoms 3"],
    amplitudes: Float[Array, "n_atoms n_gaussians"],
    variances: Float[Array, "n_atoms n_gaussians"],
    images: Float[Array, "n_images y x"],
    image_config: cxs.BasicImageConfig,
    transfer_theories: cxs.ContrastTransferTheory,
    pose_search: HierarchicalSO3GridSearch,
    n_images_in_parallel: int,
) -> cxs.QuaternionPose:
    volume = cxs.GaussianMixtureVolume(walker, amplitudes, variances)
    pose_search_vmap = eqx.filter_vmap(
        lambda img, ic, tf: pose_search(volume, img, ic, tf),
        in_axes=(0, eqx.if_array(0), 0),
    )

    return filter_bmap(
        pose_search_vmap,
        xs=(images, image_config, transfer_theories),
        batch_size=n_images_in_parallel,
    )



from typing import TypedDict

import numpy as np
from cryojax.simulator import BasicImageConfig, ContrastTransferTheory, EulerAnglePose
from jaxtyping import Float


class ParticleParameterInfo(TypedDict):
    """Parameters for a particle stack from RELION."""

    image_config: BasicImageConfig
    pose: EulerAnglePose
    transfer_theory: ContrastTransferTheory


class ParticleStackInfo(TypedDict):
    """Particle stack info from RELION."""

    parameters: ParticleParameterInfo | None
    images: Float[np.ndarray, "... y_dim x_dim"]



import mdtraj
import pytest

# from jaxtyping import install_import_hook
# with install_import_hook("cryojax_ensemble_optimization", "typeguard.typechecked"):
from cryojax_eo.io import (
    read_walkers_from_pdbs,
)


@pytest.fixture
def path_to_atomic_models(sample_path_to_pdb1, sample_path_to_pdb2):
    return [sample_path_to_pdb1, sample_path_to_pdb2]


@pytest.mark.parametrize(
    "loads_b_factors",
    [
        True,
        False,
    ],
)
def test_load_atomic_models_from_pdb(path_to_atomic_models, loads_b_factors):
    """
    Test loading all atoms from the atomic models.
    """
    walkers, amplitudes, variances = read_walkers_from_pdbs(
        path_to_atomic_models,
        loads_b_factors=loads_b_factors,
    )

    atom_positions = mdtraj.load(
        path_to_atomic_models[0],
    )
    atom_positions = atom_positions.xyz[0]

    assert walkers.shape == (len(path_to_atomic_models), atom_positions.shape[0], 3)
    assert amplitudes.shape == (atom_positions.shape[0], 5)
    assert variances.shape == (atom_positions.shape[0], 5)


def test_inconsistent_pdbs(sample_path_groel_pdb, sample_path_to_pdb1):
    """
    Test that an error is raised when loading atomic models from PDB files with
    inconsistent atomic composition.
    """
    with pytest.raises(ValueError):
        read_walkers_from_pdbs(
            [sample_path_groel_pdb, sample_path_to_pdb1],
            loads_b_factors=False,
        )



import os
import pathlib
from typing import Any

import equinox as eqx
import jax.numpy as jnp
import mdtraj
import optax
from jaxtyping import Array, Float, Int
from mdtraj.formats import XTCTrajectoryFile
from tqdm import tqdm

from ...ensemble_optimization._prior_projection._molecular_dynamics.openmm import (
    SteeredMDSimulator,
)
from .._cross_corelation.optimizer import SteepestDescWalkerFlexibleFitting


class FlexibleFittingPipeline(eqx.Module):
    """
    Ensemble refinement pipeline using OpenMM for molecular dynamics simulation.
    """

    prior_projector: SteeredMDSimulator
    likelihood_optimizer: SteepestDescWalkerFlexibleFitting
    n_steps: int
    reference_structure: mdtraj.Trajectory
    atom_indices_for_opt: Int[Array, " n_atoms_for_opt"]

    def __init__(
        self,
        prior_projector: SteeredMDSimulator,
        likelihood_optimizer: SteepestDescWalkerFlexibleFitting,
        n_steps: int,
        ref_structure_for_alignment: mdtraj.Trajectory,
        atom_indices_for_opt: Int[Array, " n_atoms_for_opt"],
    ):
        self.prior_projector = prior_projector
        self.likelihood_optimizer = likelihood_optimizer
        self.n_steps = n_steps
        self.reference_structure = ref_structure_for_alignment
        self.atom_indices_for_opt = atom_indices_for_opt

    def run(
        self,
        initial_walker: Float[Array, "n_atoms 3"],
        reference_volume: Float[Array, "n_pixels n_pixels n_pixels"],
        bias_constant_scheduler: optax.ScalarOrSchedule,
        *,
        output_directory: str | pathlib.Path,
        initial_state_for_projector: Any = None,
    ) -> tuple[Float[Array, "n_atoms 3"], Any]:
        # print("Initializing projetor...")
        md_state = self.prior_projector.initialize(initial_state_for_projector)
        # print("Projector initialized.")

        reference_structure = self.reference_structure
        walker = initial_walker.copy()

        # print("Preparing writers for output...")
        writer = XTCTrajectoryFile(os.path.join(output_directory, "traj_walker.xtc"), "w")
        # print("Writers prepared.")

        # print("Aligning walker to reference structure...")
        walker = _align_walkers_to_reference(
            walker, reference_structure, self.atom_indices_for_opt
        )
        # print("Walkers aligned.")

        progress_bar = tqdm(range(self.n_steps), desc="Flexible Fitting", leave=True)
        for i in progress_bar:
            """
            if stride_for_pose is True:
                new_dataset = pose_estimation(walker)
                dataloader = create_dataloader...
            """

            # print("Likelihood Optimization: ")
            loss, tmp_walker = self.likelihood_optimizer(
                walker[self.atom_indices_for_opt, :],
                reference_volume,
            )

            ref_walker = walker.at[self.atom_indices_for_opt, :].set(tmp_walker)
            ref_walker.block_until_ready()
            # print("Likelihood Optimization done.")

            # print("Prior Projection: ")
            walker, md_state = self.prior_projector(
                ref_walker, md_state, bias_constant_scheduler(i)
            )

            walker = _align_walkers_to_reference(
                walker, reference_structure, self.atom_indices_for_opt
            )

            reference_structure = mdtraj.Trajectory(
                ref_walker / 10.0,
                topology=reference_structure.topology,
            )

            # print("Write trajectory to files...")
            writer.write(walker / 10.0)

            progress_bar.set_description(f"Flexible Fitting (C.C: {1 - loss:.4f})")

        writer.close()
        mdtraj.Trajectory(
            xyz=walker / 10.0,
            topology=self.reference_structure.topology,
        ).save_pdb(os.path.join(output_directory, "final_walker.pdb"))

        return walker, md_state


def _align_walkers_to_reference(
    walker: Float[Array, "n_atoms 3"],
    reference_structure: mdtraj.Trajectory,
    atom_indices: Int[Array, " n_atoms_for_opt"],
) -> Float[Array, "n_atoms 3"]:
    """
    Align the walker to the reference structure.
    """
    # Convert walker to mdtraj format
    walkers_mdtraj = mdtraj.Trajectory(
        xyz=walker / 10.0,  # Convert to nm
        topology=reference_structure.topology,
    ).superpose(
        reference_structure,
        frame=0,
        atom_indices=atom_indices,
    )
    return jnp.array(walkers_mdtraj.xyz[0]) * 10.0  # Convert back to Angstroms



"""
Copyright (c) 2023 ghcollin

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

"""
These functions were ported from
https://github.com/ghcollin/healjax/tree/master

Please see Copyright notice above for more information.
"""

import jax
import jax.numpy as jnp


def hp_to_zphi_polar(nside, bighp, x, y):
    zfactor, x, y = jax.lax.cond(
        jnp.logical_and(issouthpolar(bighp), x + y < nside),
        lambda: (-1.0, nside - y, nside - x),
        lambda: (1.0, x, y),
    )

    phi_t_flag = jnp.logical_or(y != nside, x != nside)
    phi_t = phi_t_flag * (jnp.pi * (nside - y) / (2.0 * ((nside - x) + (nside - y))))

    vv = jnp.where(
        phi_t < jnp.pi / 4,
        jnp.fabs(jnp.pi * (nside - x) / ((2.0 * phi_t - jnp.pi) * nside) / jnp.sqrt(3.0)),
        jnp.fabs(jnp.pi * (nside - y) / (2.0 * phi_t * nside) / jnp.sqrt(3.0)),
    )

    z = (1 - vv) * (1 + vv)
    rad = jnp.sqrt(1.0 + z) * vv

    z = z * zfactor

    # // The big healpix determines the phi offset
    phi = jnp.where(
        issouthpolar(bighp),
        jnp.pi / 2.0 * (bighp - 8) + phi_t,
        jnp.pi / 2.0 * bighp + phi_t,
    )
    phi = jnp.mod(phi, 2 * jnp.pi)
    return z, phi, rad


def hp_to_zphi_equator(nside, bighp, x, y):
    x = x / nside
    y = y / nside

    bighp, zoff, phioff = jax.lax.cond(
        bighp <= 3,
        lambda: (bighp, 0.0, 1.0),  # // north
        lambda: jax.lax.cond(
            bighp <= 7,
            lambda: (bighp - 4, -1.0, 0.0),  # // equator
            lambda: (bighp - 8, -2.0, 1.0),  # // south
        ),
    )

    z = 2.0 / 3.0 * (x + y + zoff)
    phi = jnp.pi / 4 * (x - y + phioff + 2 * bighp)
    phi = jnp.mod(phi, 2 * jnp.pi)
    rad = jnp.sqrt(
        jnp.maximum(0.0, 1 - jnp.square(z))
    )  # This sqrt can cause spirious NaN errors in debug mode, so we clip the input
    return z, phi, rad


def hp_to_zphi(nside, bighp, xp, yp, dx, dy):
    # // this is x,y position in the healpix reference frame
    x = xp + dx
    y = yp + dy

    polar_routine = jnp.logical_or(
        jnp.logical_and(isnorthpolar(bighp), x + y > nside),
        jnp.logical_and(issouthpolar(bighp), x + y < nside),
    )

    return jax.lax.cond(
        polar_routine, hp_to_zphi_polar, hp_to_zphi_equator, nside, bighp, x, y
    )


def zphi2xyz(z, phi, rad):
    x = rad * jnp.cos(phi)
    y = rad * jnp.sin(phi)
    return x, y, z


def xyz2radec(x, y, z):
    ra = jnp.mod(jnp.arctan2(y, x), 2 * jnp.pi)
    dec = jnp.where(
        jnp.fabs(z) > 0.9, jnp.pi / 2 - jnp.arctan2(jnp.hypot(x, y), z), jnp.arcsin(z)
    )
    return ra, dec


def healpixl_nested_to_xy(nside, hp):
    ns2 = nside * nside
    bighp = (hp / ns2).astype(hp.dtype)

    def loop_body(i, carry):
        index0, xc, yc = carry
        new_x = jnp.bitwise_or(xc, jnp.left_shift(jnp.bitwise_and(index0, 0x1), i))
        index1 = jnp.right_shift(index0, 1)
        new_y = jnp.bitwise_or(yc, jnp.left_shift(jnp.bitwise_and(index1, 0x1), i))
        index2 = jnp.right_shift(index1, 1)
        return index2, new_x, new_y

    index = jnp.fmod(hp, ns2)
    x_dtype = hp.dtype
    _, x, y = jax.lax.fori_loop(
        0,
        8 * x_dtype.itemsize // 2,
        loop_body,
        (index, jnp.array(0).astype(x_dtype), jnp.array(0).astype(x_dtype)),
    )

    return bighp, x, y


def isnorthpolar(bighealpix):
    return bighealpix <= 3


def issouthpolar(bighealpix):
    return bighealpix >= 8


##############
# API funcs
##############


def vec2ang_radec(x, y, z):
    return xyz2radec(x, y, z)


def vec2ang(x, y, z):
    ra, dec = vec2ang_radec(x, y, z)
    phi = ra
    theta = jnp.pi / 2 - dec
    return theta, phi


def _pix2ang(nside, hp, dx=None, dy=None):
    dx = 0.5 if dx is None else dx
    dy = 0.5 if dy is None else dy
    return vec2ang(
        *zphi2xyz(*hp_to_zphi(nside, *healpixl_nested_to_xy(nside, hp), dx, dy))
    )


@jax.jit
def pix2ang(nside, hp_indices):
    return jax.vmap(lambda x: _pix2ang(nside, x), in_axes=0)(hp_indices)



#!/usr/bin/env python3
import argparse
import datetime
import logging
import os
import sys
from pathlib import Path

import jax.numpy as jnp
import yaml

from ..internal._config_validators.gmm_fit_config import GMMFitConfig
from ..internal._config_validators.utils import _validate_files_with_type
from ..utils._gmm_fitting import make_gmm_model_from_atomic_model


def add_args(parser):
    parser.add_argument(
        "--config", type=str, default=None, help="Path to the config (yaml) file"
    )
    # add arguments for input and output files
    parser.add_argument(
        "-o", "--output_file", type=str, default=None, help="Path to the output file"
    )
    parser.add_argument(
        "-i", "--input_pdb", type=str, default=None, help="Path to the reference file"
    )
    return parser


def mkbasedir(out):
    if not os.path.exists(out):
        try:
            os.makedirs(out)
        except (FileExistsError, PermissionError):
            raise ValueError("Output path does not exist and cannot be created.")
    return


def warnexists(out):
    if os.path.exists(out):
        Warning(f"Warning: {out} already exists. Overwriting.")


def main(args):
    config_file = args.config
    input_pdb = args.input_pdb
    output_file = Path(args.output_file)

    _validate_files_with_type(input_pdb, [".pdb"])
    if config_file is None:
        config = GMMFitConfig()
        config_file = "default_config.yaml"
        config_dict = dict(config.model_dump())
    else:
        with open(config_file) as f:
            config_dict = yaml.safe_load(f)
            config = GMMFitConfig(**config_dict)

    basedir = output_file.parent
    warnexists(basedir)
    mkbasedir(basedir)
    print(f"A copy of the config file and the log will be written to {basedir}")
    sys.stdout.flush()

    # make copy of config to output_path

    logger = logging.getLogger()
    logger_fname = datetime.datetime.now().strftime("%Y-%m-%d-%H")
    logger_fname = os.path.join(basedir, logger_fname + ".log")
    fhandler = logging.FileHandler(filename=logger_fname, mode="a")
    formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    fhandler.setFormatter(formatter)
    logger.addHandler(fhandler)
    logger.setLevel(logging.INFO)

    config_fname = os.path.basename(config_file)
    with open(os.path.join(basedir, config_fname), "w") as f:
        yaml.dump(config_dict, f, default_flow_style=False)

    logging.info(
        f"A copy of the used config file has been written "
        f"to {os.path.join(basedir, config_fname)}"
    )

    logging.info("Simulating particle stack...")

    fitted_gmm = make_gmm_model_from_atomic_model(
        pdb_file=input_pdb,
        box_size=config.box_size,
        voxel_size=config.voxel_size,
        fit_selection_string=config.fit_selection_string,
        init_amp=config.init_log_amp,
        init_var=config.init_log_var,
        n_gaussians_per_bead=config.n_gaussians_per_bead,
        atol=config.atol,
        rtol=config.rtol,
        max_steps=config.max_steps,
    )

    jnp.savez(
        output_file,
        positions=fitted_gmm.positions,
        amplitudes=fitted_gmm.amplitudes,
        variances=fitted_gmm.variances,
    )

    logging.info("Simulation complete.")
    logging.info(f"Output written to {basedir}")

    return


def main_cli():
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=yaml.dump(GMMFitConfig.model_json_schema(), indent=4),
    )

    main(add_args(parser).parse_args())


if __name__ == "__main__":
    main_cli()



from pydantic import (
    BaseModel,
    Field,
    PositiveFloat,
    PositiveInt,
)


class GMMFitConfig(BaseModel, extra="forbid"):
    box_size: PositiveInt = Field(
        default=128,
        description="Size of the box in pixels."
        + "This is used to compute the density used to fit the GMM",
    )
    voxel_size: PositiveFloat = Field(
        default=2.0,
        description="Size of the voxel in Angstroms."
        + "This is used to compute the density used to fit the GMM",
    )
    fit_selection_string: str = Field(
        default='name "C2"',
        description="Selection string to select the atoms used to define the GMM."
        + "This is used to select the atoms from the reference file.",
    )
    init_log_amp: float = Field(
        default=40.0,
        description="Initial log amplitude for the GMM fitting."
        + "This is used to initialize the GMM fitting.",
    )
    init_log_var: float = Field(
        default=1.0,
        description="Initial log variance for the GMM fitting."
        + "This is used to initialize the GMM fitting.",
    )
    n_gaussians_per_bead: PositiveInt = Field(
        default=1,
        description="Number of Gaussians per bead in the GMM fitting."
        + "This is used to define the number of Gaussians per bead in the GMM fitting.",
    )
    atol: float = Field(
        default=1e-3,
        description="Absolute tolerance for the GaussNewton Algorithm used to fit the GMM."  # noqa
        + "This is used to define the convergence criteria for the GMM fitting.",
    )
    rtol: float = Field(
        default=1e-3,
        description="Relative tolerance for the GaussNewton Algorithm used to fit the GMM."  # noqa
        + "This is used to define the convergence criteria for the GMM fitting.",
    )
    max_steps: PositiveInt = Field(
        default=500,
        description="Maximum number of steps for the GaussNewton Algorithm used to fit the GMM."  # noqa
        + "This is used to define the maximum number of steps for the GMM fitting.",
    )



from . import (
    dataset as dataset,
    ensemble_optimization as ensemble_optimization,
    flexible_fitting as flexible_fitting,
    io as io,
    simulator as simulator,
    typing as typing,
    utils as utils,
)
from .commands import (
    run_ensemble_optimization_with_md as run_ensemble_optimization_with_md,
)
from .cryojax_eo_version import __version__ as __version__
from .internal import load_config as load_config



import jax.numpy as jnp
import jax.random as jr
import pytest
from cryojax.rotations import SO3

from cryojax_eo.utils import rigid_align_positions


@pytest.mark.parametrize("n_atoms", [1000, 10000, 20000])
def test_rigid_align_positions(n_atoms):
    key = jr.key(0)
    key_rot, key_trans, key_points = jr.split(key, 3)

    # Rantom transformation
    random_rotation = SO3.sample_uniform(key_rot).as_matrix()
    random_translation = jr.normal(key_trans, (3,)) * 10.0

    random_points = jr.normal(key_points, (n_atoms, 3)) * 10.0
    random_points -= jnp.mean(random_points, axis=0, keepdims=True)  # Center the points

    transformed_points = (random_points + random_translation) @ random_rotation.T

    aligned_points, rot_matrix, displacement = rigid_align_positions(
        transformed_points, random_points
    )

    print(f"Test with {n_atoms} atoms:")
    print(random_translation, displacement)
    assert jnp.allclose(
        aligned_points, random_points, atol=1e-5
    ), "Aligned points do not match reference points"
    assert jnp.allclose(
        rot_matrix @ random_rotation, jnp.eye(3), atol=1e-5
    ), "Rotation matrices do not match"
    assert jnp.allclose(displacement, -random_translation), "Translations do not match"
